# -*- coding: UTF-8 -*-
# @Time    : 2019-06-08 15:51
# @File    : lsp.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
from antgo.dataflow.dataset import *
import scipy.io as io
import os
import numpy as np
import cv2
import time
from antgo.framework.helper.fileio.file_client import *

__all__ = ['LSP']

class LSP(Dataset):
    def __init__(self, train_or_test, dir=None, ext_params=None):
        super(LSP, self).__init__(train_or_test, dir, ext_params)
        self.class_name = [
            'Right ankle',
            'Right knee',
            'Right hip',
            'Left hip',
            'Left knee',
            'Left ankle',
            'Right wrist',
            'Right elbow',
            'Right shoulder',
            'Left shoulder',
            'Left elbow',
            'Left wrist',
            'Neck',
            'Head top'
        ]

        if os.environ.get('LOCAL_RANK', 0) == 0:
            if not os.path.exists(os.path.join(self.dir, 'lsp','joints.mat')):
                # 数据集不存在，需要重新下载，并创建标记
                if not os.path.exists(self.dir):
                    os.makedirs(self.dir)

                os.makedirs(os.path.join(self.dir, 'lsp'), exist_ok=True)
                os.makedirs(os.path.join(self.dir, 'lspet', 'images'), exist_ok=True)

                ali = AliBackend()
                ali.download('ali:///dataset/lsp/lsp_dataset.zip', os.path.join(self.dir, 'lsp'))
                # ali.download('ali:///dataset/lsp/lspet_dataset.zip', os.path.join(self.dir, 'lspet'))
                ali.download('ali:///dataset/lsp/hr-lspet.zip', self.dir)   # 这是lspet的高精集合
                
                os.system(f'cd {os.path.join(self.dir, "lsp")} && unzip lsp_dataset.zip')
                # os.system(f'cd {os.path.join(self.dir, "lspet")} && unzip lspet_dataset.zip')            
                os.system(f'cd {self.dir} && unzip hr-lspet.zip && mv hr-lspet/*.png lspet/images && mv hr-lspet/* lspet/')
                os.system('touch DATASET_IS_READY')
            else:
                # 数据集存在，创建标记
                if not os.path.exists('DATASET_IS_READY'):
                    os.system('touch DATASET_IS_READY')
        else:
            while True:
                # 等待直到存在指定文件
                if os.path.exists('DATASET_IS_READY'):
                    break
                time.sleep(5)

        self.dataset = []
        # lsp (1000 train + 1000 test)
        lsp_matr = io.loadmat(os.path.join(self.dir, 'lsp','joints.mat'))['joints']
        train_idx = list(range(0,2000,2))
        test_idx = list(range(1,2000,2))
        if self.train_or_test == 'train':
            # path, 14x2, visible
            self.dataset = [(os.path.join(self.dir, 'lsp', 'images', 'im%04d.jpg'%(idx+1)), np.transpose(lsp_matr[:2,:,idx], [1,0]), 1-lsp_matr[2,:,idx]) for idx in train_idx]
        else:
            self.dataset = [(os.path.join(self.dir, 'lsp', 'images', 'im%04d.jpg'%(idx+1)), np.transpose(lsp_matr[:2,:,idx], [1,0]), 1-lsp_matr[2,:,idx]) for idx in test_idx]

        # lspet
        lspet_matr = io.loadmat(os.path.join(self.dir, 'lspet', 'joints.mat'))['joints']
        if self.train_or_test == 'train':
            lspet_dataset = []
            count = 0
            for idx in range(10000):
                if not os.path.exists(os.path.join(self.dir, 'lspet', 'images', 'im%05d.png'%(idx+1))):
                    continue

                lspet_dataset.append((
                    os.path.join(self.dir, 'lspet', 'images', 'im%05d.png'%(idx+1)),
                    lspet_matr[:,:2,count],
                    lspet_matr[:,2,count],
                ))
                count += 1
            
            self.dataset.extend(lspet_dataset)

    @property
    def size(self):
        return len(self.dataset)

    def at(self, id):
        
        image_file, joints2d, visible = self.dataset[id]
        image = cv2.imread(image_file)
        visible_pos = np.where(visible == 1)
        bboxes = np.array([[
          np.min(joints2d[visible_pos, 0]), 
          np.min(joints2d[visible_pos, 1]), 
          np.max(joints2d[visible_pos, 0]), 
          np.max(joints2d[visible_pos, 1])
        ]])
        anno = {
            'bboxes': bboxes.astype(np.float32),
            'labels': np.zeros((1), dtype=np.int32),
            'joints2d': np.expand_dims(joints2d, 0).astype(np.float32),
            'joints_vis': np.expand_dims(visible, 0).astype(np.int32)
        }
        return (image, anno)

    def split(self, split_params={}, split_method=''):
        raise NotImplementedError

# lsp = LSP('test', '/root/workspace/dataset/lsp')
# crjo = ConvertRandomObjJointsAndOffset(input_size=(128,128), heatmap_size=(16,16), num_joints=14)

# for i in range(lsp.size):
#     data = lsp.sample(i)
#     data = crjo(data)
    
#     image = data['image']
#     joints2d = data['joints2d']
#     joints_vis = data['joints_vis']
    
#     # for joint_i, (x,y) in enumerate(joints2d[0]):
#     #     x, y = int(x), int(y)
        
#     #     if joints_vis[0][joint_i]:
#     #         cv2.circle(image, (x, y), radius=2, color=(0,0,255), thickness=1)

#     # cv2.imwrite(f'./aabb_{i}.png', image)
#     print(i)


import copy
import os.path as osp
from abc import ABCMeta, abstractmethod
from os import PathLike
from typing import List
from io import StringIO
import numpy as np
from numbers import Number
from antgo.dataflow.dataset import Dataset


def expanduser(path):
    if isinstance(path, (str, PathLike)):
        return osp.expanduser(path)
    else:
        return path

def list_from_file(filename,
                   prefix='',
                   offset = 0,
                   max_num= 0,
                   encoding = 'utf-8'):
    """Load a text file and parse the content as a list of strings.

    Note:
        In v1.3.16 and later, ``list_from_file`` supports loading a text file
        which can be storaged in different backends and parsing the content as
        a list for strings.

    Args:
        filename (str): Filename.
        prefix (str): The prefix to be inserted to the beginning of each item.
        offset (int): The offset of lines.
        max_num (int): The maximum number of lines to be read,
            zeros and negatives mean no limitation.
        encoding (str): Encoding used to open the file. Default utf-8.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Default: None.

    Examples:
        >>> list_from_file('/path/of/your/file')  # disk
        ['hello', 'world']
        >>> list_from_file('s3://path/of/your/file')  # ceph or petrel
        ['hello', 'world']

    Returns:
        list[str]: A list of strings.
    """
    cnt = 0
    item_list = []
    value_buf = None
    with open(filename, encoding=encoding) as f:
        value_buf = f.read()

    with StringIO(value_buf) as f:
        for _ in range(offset):
            f.readline()
        for line in f:
            if 0 < max_num <= cnt:
                break
            item_list.append(prefix + line.rstrip('\n\r'))
            cnt += 1
    return item_list

def accuracy_numpy(pred, target, topk=(1, ), thrs=0.):
    if isinstance(thrs, Number):
        thrs = (thrs, )
        res_single = True
    elif isinstance(thrs, tuple):
        res_single = False
    else:
        raise TypeError(
            f'thrs should be a number or tuple, but got {type(thrs)}.')

    res = []
    maxk = max(topk)
    num = pred.shape[0]

    static_inds = np.indices((num, maxk))[0]
    pred_label = pred.argpartition(-maxk, axis=1)[:, -maxk:]
    pred_score = pred[static_inds, pred_label]

    sort_inds = np.argsort(pred_score, axis=1)[:, ::-1]
    pred_label = pred_label[static_inds, sort_inds]
    pred_score = pred_score[static_inds, sort_inds]

    for k in topk:
        correct_k = pred_label[:, :k] == target.reshape(-1, 1)
        res_thr = []
        for thr in thrs:
            # Only prediction values larger than thr are counted as correct
            _correct_k = correct_k & (pred_score[:, :k] > thr)
            _correct_k = np.logical_or.reduce(_correct_k, axis=1)
            res_thr.append((_correct_k.sum() * 100. / num))
        if res_single:
            res.append(res_thr[0])
        else:
            res.append(res_thr)
    return res


def accuracy_numpy(pred, target, topk=(1, ), thrs=0.):
    if isinstance(thrs, Number):
        thrs = (thrs, )
        res_single = True
    elif isinstance(thrs, tuple):
        res_single = False
    else:
        raise TypeError(
            f'thrs should be a number or tuple, but got {type(thrs)}.')

    res = []
    maxk = max(topk)
    num = pred.shape[0]

    static_inds = np.indices((num, maxk))[0]
    pred_label = pred.argpartition(-maxk, axis=1)[:, -maxk:]
    pred_score = pred[static_inds, pred_label]

    sort_inds = np.argsort(pred_score, axis=1)[:, ::-1]
    pred_label = pred_label[static_inds, sort_inds]
    pred_score = pred_score[static_inds, sort_inds]

    for k in topk:
        correct_k = pred_label[:, :k] == target.reshape(-1, 1)
        res_thr = []
        for thr in thrs:
            # Only prediction values larger than thr are counted as correct
            _correct_k = correct_k & (pred_score[:, :k] > thr)
            _correct_k = np.logical_or.reduce(_correct_k, axis=1)
            res_thr.append((_correct_k.sum() * 100. / num))
        if res_single:
            res.append(res_thr[0])
        else:
            res.append(res_thr)
    return res


def accuracy(pred, target, topk=1, thrs=0.):
    """Calculate accuracy according to the prediction and target.

    Args:
        pred (torch.Tensor | np.array): The model prediction.
        target (torch.Tensor | np.array): The target of each prediction
        topk (int | tuple[int]): If the predictions in ``topk``
            matches the target, the predictions will be regarded as
            correct ones. Defaults to 1.
        thrs (Number | tuple[Number], optional): Predictions with scores under
            the thresholds are considered negative. Default to 0.

    Returns:
        torch.Tensor | list[torch.Tensor] | list[list[torch.Tensor]]: Accuracy
            - torch.Tensor: If both ``topk`` and ``thrs`` is a single value.
            - list[torch.Tensor]: If one of ``topk`` or ``thrs`` is a tuple.
            - list[list[torch.Tensor]]: If both ``topk`` and ``thrs`` is a \
              tuple. And the first dim is ``topk``, the second dim is ``thrs``.
    """
    assert isinstance(topk, (int, tuple))
    if isinstance(topk, int):
        topk = (topk, )
        return_single = True
    else:
        return_single = False

    if not isinstance(pred, np.ndarray):
        pred = pred.cpu().numpy()
    
    if not isinstance(target, np.ndarray):
        target = target.cpu().numpy()

    res = accuracy_numpy(pred, target, topk, thrs)
    return res[0] if return_single else res


class ClsDataset(Dataset, metaclass=ABCMeta):
    """Base dataset.

    Args:
        data_prefix (str): the prefix of data path
        pipeline (list): a list of dict, where each element represents
            a operation defined in `mmcls.datasets.pipelines`
        ann_file (str | None): the annotation file. When ann_file is str,
            the subclass is expected to read from the ann_file. When ann_file
            is None, the subclass is expected to read according to data_prefix
        test_mode (bool): in train mode or test mode
    """

    CLASSES = None

    def __init__(self,
                 train_or_test='train',
                 dir=None,
                 params={'data_prefix': '', 'ann_file': None, 'classes': None}):
        super(ClsDataset, self).__init__(train_or_test, dir, params)
        self.data_prefix = expanduser(params['data_prefix'])
        # self.pipeline = Compose(pipeline)
        self.CLASSES = self.get_classes(params['classes'])
        self.ann_file = expanduser(params['ann_file'])
        self.test_mode = train_or_test != 'train'
        self.data_infos = self.load_annotations()

    @abstractmethod
    def load_annotations(self):
        pass

    @property
    def class_to_idx(self):
        """Map mapping class name to class index.

        Returns:
            dict: mapping from class name to class index.
        """

        return {_class: i for i, _class in enumerate(self.CLASSES)}

    def get_gt_labels(self):
        """Get all ground-truth labels (categories).

        Returns:
            np.ndarray: categories for all images.
        """

        gt_labels = np.array([data['gt_label'] for data in self.data_infos])
        return gt_labels

    def get_cat_ids(self, idx: int) -> List[int]:
        """Get category id by index.

        Args:
            idx (int): Index of data.

        Returns:
            cat_ids (List[int]): Image category of specified index.
        """

        return [int(self.data_infos[idx]['gt_label'])]

    def prepare_data(self, idx):
        results = copy.deepcopy(self.data_infos[idx])
        return results

    @property
    def size(self):
        return len(self.data_infos)

    def at(self, idx):
        return None, self.prepare_data(idx)
       
    def data_pool(self):
        for index in range(self.size):
            yield self.at(index)

    def get_ann_info(self, idx):
        return self.prepare_data(idx)

    @classmethod
    def get_classes(cls, classes=None):
        """Get class names of current dataset.

        Args:
            classes (Sequence[str] | str | None): If classes is None, use
                default CLASSES defined by builtin dataset. If classes is a
                string, take it as a file name. The file contains the name of
                classes where each line contains one class name. If classes is
                a tuple or list, override the CLASSES defined by the dataset.

        Returns:
            tuple[str] or list[str]: Names of categories of the dataset.
        """
        if classes is None:
            return cls.CLASSES

        if isinstance(classes, str):
            # take it as a file path
            class_names = list_from_file(expanduser(classes))
        elif isinstance(classes, (tuple, list)):
            class_names = classes
        elif type(classes) == int:
            class_names = [(str(i),i) for i in range(classes)]
        else:
            raise ValueError(f'Unsupported type {type(classes)} of classes.')

        return class_names

    def evaluate(self,
                 results,
                 metric='accuracy',
                 metric_options=None,
                 indices=None,
                 logger=None):
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
                Default value is `accuracy`.
            metric_options (dict, optional): Options for calculating metrics.
                Allowed keys are 'topk', 'thrs' and 'average_mode'.
                Defaults to None.
            indices (list, optional): The indices of samples corresponding to
                the results. Defaults to None.
            logger (logging.Logger | str, optional): Logger used for printing
                related information during evaluation. Defaults to None.
        Returns:
            dict: evaluation results
        """
        if metric_options is None:
            metric_options = {'topk': (1, 5)}
        if isinstance(metric, str):
            metrics = [metric]
        else:
            metrics = metric
        allowed_metrics = [
            'accuracy', 'precision', 'recall', 'f1_score'
        ]
        eval_results = {}
        results = np.vstack([v['pred'] for v in results])
        gt_labels = self.get_gt_labels()
        if indices is not None:
            gt_labels = gt_labels[indices]
        num_imgs = len(results)
        assert len(gt_labels) == num_imgs, 'dataset testing results should '\
            'be of the same length as gt_labels.'

        invalid_metrics = set(metrics) - set(allowed_metrics)
        if len(invalid_metrics) != 0:
            raise ValueError(f'metric {invalid_metrics} is not supported.')

        topk = metric_options.get('topk', (1, 5))
        thrs = metric_options.get('thrs')
        average_mode = metric_options.get('average_mode', 'macro')

        if 'accuracy' in metrics:
            if thrs is not None:
                acc = accuracy(results, gt_labels, topk=topk, thrs=thrs)
            else:
                acc = accuracy(results, gt_labels, topk=topk)
            if isinstance(topk, tuple):
                eval_results_ = {
                    f'accuracy_top-{k}': a
                    for k, a in zip(topk, acc)
                }
            else:
                eval_results_ = {'accuracy': acc}
            if isinstance(thrs, tuple):
                for key, values in eval_results_.items():
                    eval_results.update({
                        f'{key}_thr_{thr:.2f}': value.item()
                        for thr, value in zip(thrs, values)
                    })
            else:
                eval_results.update(
                    {k: v.item()
                     for k, v in eval_results_.items()})

        return eval_results



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import warnings

import paddle
from paddle.regularizer import L2Decay



class _RequiredParameter(object):
    """Singleton class representing a required parameter for an Optimizer."""

    def __repr__(self):
        return "<required parameter>"


required = _RequiredParameter()


def update_parameters(parameters, lr, weight_decay):
    parameters_list = list()
    if parameters is not None:
        for items in parameters:
            if isinstance(items, dict):
                params = items["params"]
                if "lr" in items:
                    for p in params:
                        p.optimize_attr["learning_rate"] = items[
                            "lr"] / lr * p.optimize_attr["learning_rate"]
                if "weight_decay" in items:
                    for p in params:
                        if isinstance(items["weight_decay"], (float, int)):
                            p.regularizer = L2Decay(items["weight_decay"])
                        else:
                            p.regularizer = weight_decay
                for p in params:
                    print(p.regularizer)
                parameters_list.extend(params)
            else:
                parameters_list.append(items)
    return parameters_list


class Momentum(paddle.optimizer.Momentum):
    def __init__(self,
                 params,
                 lr=0.001,
                 momentum=0.0,
                 dampening=0,
                 weight_decay=0.0,
                 nesterov=False):
        assert dampening == 0, "The dampening must be 0 in Momentum!"
        parameters_list = update_parameters(params, lr, weight_decay)
        super().__init__(
            learning_rate=lr,
            momentum=momentum,
            parameters=parameters_list,
            use_nesterov=nesterov,
            weight_decay=weight_decay,
            grad_clip=None,
            name=None)

        defaults = dict(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov)
        self.defaults = defaults

        self.state = defaultdict(dict)
        self.param_groups = []

        param_groups = list(params)
        if len(param_groups) == 0:
            raise ValueError("optimizer got an empty parameter list")
        if not isinstance(param_groups[0], dict):
            param_groups = [{'params': param_groups}]

        for param_group in param_groups:
            self.add_param_group(param_group)

    def add_param_group(self, param_group):
        assert isinstance(param_group, dict), "param group must be a dict"

        params = param_group['params']
        if isinstance(params, paddle.Tensor):
            param_group['params'] = [params]
        elif isinstance(params, set):
            raise TypeError(
                'optimizer parameters need to be organized in ordered collections, but '
                'the ordering of tensors in sets will change between runs. Please use a list instead.'
            )
        else:
            param_group['params'] = list(params)

        for param in param_group['params']:
            if not isinstance(param, paddle.Tensor):
                raise TypeError("optimizer can only optimize Tensors.")
            if not param.is_leaf:
                raise ValueError("can't optimize a non-leaf Tensor")

        for name, default in self.defaults.items():
            if default is required and name not in param_group:
                raise ValueError(
                    "parameter group didn't specify a value of required optimization parameter "
                    + name)
            else:
                param_group.setdefault(name, default)

        params = param_group['params']
        if len(params) != len(set(params)):
            warnings.warn(
                "optimizer contains a parameter group with duplicate parameters; "
                "in future, this will cause an error; ",
                stacklevel=3)

        param_set = set()
        for group in self.param_groups:
            param_set.update(set(group['params']))

        if not param_set.isdisjoint(set(param_group['params'])):
            raise ValueError(
                "some parameters appear in more than one parameter group")

        self.param_groups.append(param_group)

    def zero_grad(self):
        return self.clear_grad()


class Adam(paddle.optimizer.Adam):
    def __init__(self,
                 params,
                 lr=0.001,
                 betas=(0.9, 0.999),
                 eps=1e-08,
                 weight_decay=0,
                 amsgrad=False):
        parameters_list = update_parameters(params, lr, weight_decay)
        if weight_decay == 0:
            weight_decay = None
        super().__init__(
            learning_rate=lr,
            beta1=betas[0],
            beta2=betas[1],
            epsilon=eps,
            parameters=parameters_list,
            weight_decay=weight_decay,
            grad_clip=None,
            name=None,
            lazy_mode=False)

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad)
        self.defaults = defaults

        self.state = defaultdict(dict)
        self.param_groups = []

        param_groups = list(parameters_list)
        if len(param_groups) == 0:
            print(param_groups)
            raise ValueError("optimizer got an empty parameter list")
        if not isinstance(param_groups[0], dict):
            param_groups = [{'params': param_groups}]

        for param_group in param_groups:
            self.add_param_group(param_group)

    def add_param_group(self, param_group):
        assert isinstance(param_group, dict), "param group must be a dict"

        params = param_group['params']
        if isinstance(params, paddle.Tensor):
            param_group['params'] = [params]
        elif isinstance(params, set):
            raise TypeError(
                'optimizer parameters need to be organized in ordered collections, but '
                'the ordering of tensors in sets will change between runs. Please use a list instead.'
            )
        else:
            param_group['params'] = list(params)

        for param in param_group['params']:
            if not isinstance(param, paddle.Tensor):
                raise TypeError("optimizer can only optimize Tensors.")
            if not param.is_leaf:
                raise ValueError("can't optimize a non-leaf Tensor")

        for name, default in self.defaults.items():
            if default is required and name not in param_group:
                raise ValueError(
                    "parameter group didn't specify a value of required optimization parameter "
                    + name)
            else:
                param_group.setdefault(name, default)

        params = param_group['params']
        if len(params) != len(set(params)):
            warnings.warn(
                "optimizer contains a parameter group with duplicate parameters; "
                "in future, this will cause an error; ",
                stacklevel=3)

        param_set = set()
        for group in self.param_groups:
            param_set.update(set(group['params']))

        if not param_set.isdisjoint(set(param_group['params'])):
            raise ValueError(
                "some parameters appear in more than one parameter group")

        self.param_groups.append(param_group)

    def zero_grad(self):
        return self.clear_grad()


class SGD(paddle.optimizer.SGD):
    def __init__(self, params, lr, momentum=0, dampening=0, weight_decay=0, nesterov=False, *, maximize=False):
        parameters_list = update_parameters(params, lr, weight_decay)
        if weight_decay == 0:
            weight_decay = None
        super().__init__(
            learning_rate=lr,
            parameters=parameters_list,
            weight_decay=weight_decay,
            grad_clip=None,
            name=None)

        defaults = dict(
            lr=lr,
            weight_decay=weight_decay)
        self.defaults = defaults

        self.state = defaultdict(dict)
        self.param_groups = []

        param_groups = list(parameters_list)
        if len(param_groups) == 0:
            print(param_groups)
            raise ValueError("optimizer got an empty parameter list")
        if not isinstance(param_groups[0], dict):
            param_groups = [{'params': param_groups}]

        for param_group in param_groups:
            self.add_param_group(param_group)

    def add_param_group(self, param_group):
        assert isinstance(param_group, dict), "param group must be a dict"

        params = param_group['params']
        if isinstance(params, paddle.Tensor):
            param_group['params'] = [params]
        elif isinstance(params, set):
            raise TypeError(
                'optimizer parameters need to be organized in ordered collections, but '
                'the ordering of tensors in sets will change between runs. Please use a list instead.'
            )
        else:
            param_group['params'] = list(params)

        for param in param_group['params']:
            if not isinstance(param, paddle.Tensor):
                raise TypeError("optimizer can only optimize Tensors.")
            if not param.is_leaf:
                raise ValueError("can't optimize a non-leaf Tensor")

        for name, default in self.defaults.items():
            if default is required and name not in param_group:
                raise ValueError(
                    "parameter group didn't specify a value of required optimization parameter "
                    + name)
            else:
                param_group.setdefault(name, default)

        params = param_group['params']
        if len(params) != len(set(params)):
            warnings.warn(
                "optimizer contains a parameter group with duplicate parameters; "
                "in future, this will cause an error; ",
                stacklevel=3)

        param_set = set()
        for group in self.param_groups:
            param_set.update(set(group['params']))

        if not param_set.isdisjoint(set(param_group['params'])):
            raise ValueError(
                "some parameters appear in more than one parameter group")

        self.param_groups.append(param_group)

    def zero_grad(self):
        return self.clear_grad()


class LR(object):
    pass

lr_scheduler = LR()
def add_optimizer_lr_function(func):
    setattr(lr_scheduler, func.__name__, func)


class _StepLR(paddle.optimizer.lr.StepDecay):
    def __init__(self, optimizer_ft, step_size, gamma=0.1, last_epoch=- 1, verbose=False):
        lr = optimizer_ft.get_lr()
        super(_StepLR, self).__init__((float)(lr[0]) if type(lr) == list or type(lr) == tuple else (float)(lr),
                                         step_size,
                                         gamma=gamma,
                                         last_epoch=last_epoch,
                                         verbose=verbose)

    def get_lr(self):
        return [super(_StepLR, self).get_lr()]


@add_optimizer_lr_function
def StepLR(optimizer_ft, step_size, gamma=0.1, last_epoch=- 1, verbose=False):
    # learning_rate, step_size, gamma=0.1, last_epoch=- 1, verbose=False
    return _StepLR(optimizer_ft, step_size, gamma=0.1, last_epoch=- 1, verbose=False)


class ReduceOnPlateau(paddle.optimizer.lr.ReduceOnPlateau):
    def __init__(self,
                 optimizer,
                 mode='min',
                 factor=0.1,
                 patience=10,
                 threshold=0.0001,
                 threshold_mode='rel',
                 cooldown=0,
                 min_lr=0,
                 eps=1e-08,
                 verbose=False):
        super().__init__(
            learning_rate=0.01,
            mode=mode,
            factor=factor,
            patience=patience,
            threshold=threshold,
            threshold_mode=threshold_mode,
            cooldown=cooldown,
            min_lr=min_lr,
            epsilon=eps,
            verbose=verbose)
        optimizer._learning_rate = self


class CosineAnnealingLR(paddle.optimizer.lr.CosineAnnealingDecay):
    def __init__(self,
                 optimizer,
                 T_max,
                 eta_min=0,
                 last_epoch=-1,
                 verbose=False):
        super().__init__(0.01, T_max, eta_min, last_epoch, verbose)
        optimizer._learning_rate = self


class MultiStepLR(paddle.optimizer.lr.MultiStepDecay):
    def __init__(self,
                 optimizer,
                 milestones,
                 gamma=0.1,
                 last_epoch=-1,
                 verbose=False):
        super().__init__(0.01, milestones, gamma, last_epoch, verbose)
        optimizer._learning_rate = self



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod

import torch
from torchvision.ops import batched_nms 
from antgo.framework.helper.runner import BaseModule
from antgo.framework.helper.models.detectors.core.utils import filter_scores_and_topk, select_single_mlvl


class BaseDenseHead(BaseModule, metaclass=ABCMeta):
    """Base class for DenseHeads."""

    def __init__(self, init_cfg=None):
        super(BaseDenseHead, self).__init__(init_cfg)

    @abstractmethod
    def loss(self, **kwargs):
        """Compute losses of the head."""
        pass

    def forward_train(self,
                      x,
                      image_meta,
                      gt_bbox,
                      gt_class=None,
                      gt_bboxes_ignore=None,
                      **kwargs):
        """
        Args:
            x (list[Tensor]): Features from FPN.
            image_meta (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes (Tensor): Ground truth bboxes of the image,
                shape (num_gts, 4).
            gt_labels (Tensor): Ground truth labels of each box,
                shape (num_gts,).
            gt_bboxes_ignore (Tensor): Ground truth bboxes to be
                ignored, shape (num_ignored_gts, 4).
            proposal_cfg (mmcv.Config): Test / postprocessing configuration,
                if None, test_cfg would be used

        Returns:
            tuple:
                losses: (dict[str, Tensor]): A dictionary of loss components.
                proposal_list (list[Tensor]): Proposals of each image.
        """
        outs = self(x)

        if gt_class is None:
            loss_inputs = outs + (gt_bbox, image_meta)
        else:
            loss_inputs = outs + (gt_bbox, gt_class, image_meta)
        losses = self.loss(*loss_inputs, gt_bboxes_ignore=gt_bboxes_ignore)
        return losses

    def simple_test(self, feats, image_meta, rescale=True, **kwargs):
        """Test function without test-time augmentation.

        Args:
            feats (tuple[torch.Tensor]): Multi-level features from the
                upstream network, each is a 4D-tensor.
            img_metas (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[tuple[Tensor, Tensor]]: Each item in result_list is 2-tuple.
                The first item is ``bboxes`` with shape (n, 5),
                where 5 represent (tl_x, tl_y, br_x, br_y, score).
                The shape of the second tensor in the tuple is ``labels``
                with shape (n, ).
        """
        outs = self.forward(feats)
        results_list = self.get_bboxes(
            *outs, image_meta=image_meta, rescale=rescale)
        return results_list



import torch
from antgo.framework.helper.models.builder import DETECTORS, build_backbone, build_head, build_neck
from .base import BaseDetector
from .core.bbox import bbox2result
from antgo.framework.helper.runner import *


class ImageList(object):
    """
    Structure that holds a list of images (of possibly
    varying sizes) as a single tensor.
    This works by padding the images to the same size,
    and storing in a field the original sizes of each image
    """

    def __init__(self, tensors, image_sizes):
        """
        Arguments:
            tensors (tensor)
            image_sizes (list[tuple[int, int]])
        """
        self.tensors = tensors
        self.image_sizes = image_sizes

    def to(self, *args, **kwargs):
        cast_tensor = self.tensors.to(*args, **kwargs)
        return ImageList(cast_tensor, self.image_sizes)


def to_image_list(tensors, metas, size_divisible=0):
    """
    tensors can be an ImageList, a torch.Tensor or
    an iterable of Tensors. It can't be a numpy array.
    When tensors is an iterable of Tensors, it pads
    the Tensors with zeros so that they have the same
    shape
    """
    if isinstance(tensors, torch.Tensor) and size_divisible > 0:
        tensors = [tensors]

    if isinstance(tensors, ImageList):
        return tensors, metas
    elif isinstance(tensors, torch.Tensor):
        # single tensor shape can be inferred
        if tensors.dim() == 3:
            tensors = tensors[None]
        assert tensors.dim() == 4
        image_sizes = [tensor.shape[-2:] for tensor in tensors]
        return ImageList(tensors, image_sizes), metas
    elif isinstance(tensors, (tuple, list)):
        max_size = tuple(max(s) for s in zip(*[img.shape for img in tensors]))

        # TODO Ideally, just remove this and let me model handle arbitrary
        # input sizs
        if size_divisible > 0:
            import math

            stride = size_divisible
            max_size = list(max_size)
            max_size[1] = int(math.ceil(max_size[1] / stride) * stride)
            max_size[2] = int(math.ceil(max_size[2] / stride) * stride)
            max_size = tuple(max_size)

        batch_shape = (len(tensors),) + max_size
        batched_imgs = tensors[0].new(*batch_shape).zero_()
        for img, pad_img, meta in zip(tensors, batched_imgs, metas):
            pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
            meta.update({'image_shape': batch_shape[-2:]})

        image_sizes = [im.shape[-2:] for im in tensors]
        return ImageList(batched_imgs, image_sizes), metas
    else:
        raise TypeError("Unsupported type for to_image_list: {}".format(type(tensors)))


@DETECTORS.register_module()
class SingleStageDetector(BaseDetector):
    """Base class for single-stage detectors.

    Single-stage detectors directly and densely predict bounding boxes on the
    output features of the backbone+neck.
    """

    def __init__(self,
                 backbone=None,
                 neck=None,
                 bbox_head=None,
                 train_cfg=None,
                 test_cfg=None,
                 pretrained=None,
                 init_cfg=None):
        super(SingleStageDetector, self).__init__(init_cfg)
        # if pretrained:
        #     warnings.warn('DeprecationWarning: pretrained is deprecated, '
        #                   'please use "init_cfg" instead')
        #     backbone.pretrained = pretrained
        if backbone is not None:
            self.backbone = build_backbone(backbone)
        if neck is not None:
            self.neck = build_neck(neck)
        self.bbox_head = build_head(bbox_head)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg

    def extract_feat(self, img):
        """Directly extract features from the backbone+neck."""
        x = self.backbone(img)
        if self.with_neck:
            x = self.neck(x)
        return x

    def forward_dummy(self, img):
        """Used for computing network flops."""
        x = self.extract_feat(img)
        outs = self.bbox_head(x)
        return outs

    def forward_train(self,
                      image,
                      image_meta,                      
                      bboxes,
                      labels,
                      bboxes_ignore=None, **kwargs):
        """
        Args:
            image (Tensor): Input images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.
            image_meta (list[dict]): A List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                :class:`mmdet.datasets.pipelines.Collect`.
            bboxes (list[Tensor]): Each item are the truth boxes for each
                image in [tl_x, tl_y, br_x, br_y] format.
            labels (list[Tensor]): Class indices corresponding to each box
            bboxes_ignore (None | list[Tensor]): Specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        image_list, image_meta = to_image_list(image, image_meta)
        image = image_list.tensors
        super(SingleStageDetector, self).forward_train(image, image_meta)
        x = self.extract_feat(image)
        
        if bboxes is None or labels is None:
            return self.bbox_head(x)

        losses = self.bbox_head.forward_train(x, image_meta, bboxes,
                                              labels, bboxes_ignore)
        return losses
    
    def loss(self, *args, image_meta, bboxes, labels):
        losses = self.bbox_head.loss(*args, bboxes, labels, image_meta)
        return losses
    
    def simple_test(self, image, image_meta, rescale=True, **kwargs):
        """Test function without test-time augmentation.

        Args:
            image (torch.Tensor): Images with shape (N, C, H, W).
            image_meta (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[list[np.ndarray]]: BBox results of each image and classes.
                The outer list corresponds to each image. The inner list
                corresponds to each class.
        """
        image_list, image_meta = to_image_list(image, image_meta)
        image = image_list.tensors
        
        feat = self.extract_feat(image)
        results_list = self.bbox_head.simple_test(
            feat, image_meta, rescale=rescale)

        bbox_results = {
            'box': torch.stack([a for a, _ in results_list], dim=0),
            'label': torch.stack([b for _, b in results_list], dim=0)
        }
        return bbox_results


import torch.nn.functional as F


def QualityFocalLoss(
    pred_sigmoid,
    teacher_sigmoid,
    weight=None,
    beta=2.0,
    reduction="mean",
):
    pt = pred_sigmoid
    zerolabel = pt.new_zeros(pt.shape)
    loss = F.binary_cross_entropy(pred_sigmoid, zerolabel, reduction="none") * pt.pow(beta)
    pos = weight > 0

    pt = teacher_sigmoid[pos] - pred_sigmoid[pos]
    loss[pos] = F.binary_cross_entropy(pred_sigmoid[pos], teacher_sigmoid[pos], reduction="none") * pt.pow(beta)

    valid = weight >= 0
    if reduction == "mean":
        loss = loss[valid].mean()
    elif reduction == "sum":
        loss = loss[valid].sum()
    return loss



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch

from ..builder import BBOX_SAMPLERS
from .random_sampler import RandomSampler


@BBOX_SAMPLERS.register_module()
class IoUBalancedNegSampler(RandomSampler):
    """IoU Balanced Sampling.

    arXiv: https://arxiv.org/pdf/1904.02701.pdf (CVPR 2019)

    Sampling proposals according to their IoU. `floor_fraction` of needed RoIs
    are sampled from proposals whose IoU are lower than `floor_thr` randomly.
    The others are sampled from proposals whose IoU are higher than
    `floor_thr`. These proposals are sampled from some bins evenly, which are
    split by `num_bins` via IoU evenly.

    Args:
        num (int): number of proposals.
        pos_fraction (float): fraction of positive proposals.
        floor_thr (float): threshold (minimum) IoU for IoU balanced sampling,
            set to -1 if all using IoU balanced sampling.
        floor_fraction (float): sampling fraction of proposals under floor_thr.
        num_bins (int): number of bins in IoU balanced sampling.
    """

    def __init__(self,
                 num,
                 pos_fraction,
                 floor_thr=-1,
                 floor_fraction=0,
                 num_bins=3,
                 **kwargs):
        super(IoUBalancedNegSampler, self).__init__(num, pos_fraction,
                                                    **kwargs)
        assert floor_thr >= 0 or floor_thr == -1
        assert 0 <= floor_fraction <= 1
        assert num_bins >= 1

        self.floor_thr = floor_thr
        self.floor_fraction = floor_fraction
        self.num_bins = num_bins

    def sample_via_interval(self, max_overlaps, full_set, num_expected):
        """Sample according to the iou interval.

        Args:
            max_overlaps (torch.Tensor): IoU between bounding boxes and ground
                truth boxes.
            full_set (set(int)): A full set of indices of boxes。
            num_expected (int): Number of expected samples。

        Returns:
            np.ndarray: Indices  of samples
        """
        max_iou = max_overlaps.max()
        iou_interval = (max_iou - self.floor_thr) / self.num_bins
        per_num_expected = int(num_expected / self.num_bins)

        sampled_inds = []
        for i in range(self.num_bins):
            start_iou = self.floor_thr + i * iou_interval
            end_iou = self.floor_thr + (i + 1) * iou_interval
            tmp_set = set(
                np.where(
                    np.logical_and(max_overlaps >= start_iou,
                                   max_overlaps < end_iou))[0])
            tmp_inds = list(tmp_set & full_set)
            if len(tmp_inds) > per_num_expected:
                tmp_sampled_set = self.random_choice(tmp_inds,
                                                     per_num_expected)
            else:
                tmp_sampled_set = np.array(tmp_inds, dtype=np.int)
            sampled_inds.append(tmp_sampled_set)

        sampled_inds = np.concatenate(sampled_inds)
        if len(sampled_inds) < num_expected:
            num_extra = num_expected - len(sampled_inds)
            extra_inds = np.array(list(full_set - set(sampled_inds)))
            if len(extra_inds) > num_extra:
                extra_inds = self.random_choice(extra_inds, num_extra)
            sampled_inds = np.concatenate([sampled_inds, extra_inds])

        return sampled_inds

    def _sample_neg(self, assign_result, num_expected, **kwargs):
        """Sample negative boxes.

        Args:
            assign_result (:obj:`AssignResult`): The assigned results of boxes.
            num_expected (int): The number of expected negative samples

        Returns:
            Tensor or ndarray: sampled indices.
        """
        neg_inds = torch.nonzero(assign_result.gt_inds == 0, as_tuple=False)
        if neg_inds.numel() != 0:
            neg_inds = neg_inds.squeeze(1)
        if len(neg_inds) <= num_expected:
            return neg_inds
        else:
            max_overlaps = assign_result.max_overlaps.cpu().numpy()
            # balance sampling for negative samples
            neg_set = set(neg_inds.cpu().numpy())

            if self.floor_thr > 0:
                floor_set = set(
                    np.where(
                        np.logical_and(max_overlaps >= 0,
                                       max_overlaps < self.floor_thr))[0])
                iou_sampling_set = set(
                    np.where(max_overlaps >= self.floor_thr)[0])
            elif self.floor_thr == 0:
                floor_set = set(np.where(max_overlaps == 0)[0])
                iou_sampling_set = set(
                    np.where(max_overlaps > self.floor_thr)[0])
            else:
                floor_set = set()
                iou_sampling_set = set(
                    np.where(max_overlaps > self.floor_thr)[0])
                # for sampling interval calculation
                self.floor_thr = 0

            floor_neg_inds = list(floor_set & neg_set)
            iou_sampling_neg_inds = list(iou_sampling_set & neg_set)
            num_expected_iou_sampling = int(num_expected *
                                            (1 - self.floor_fraction))
            if len(iou_sampling_neg_inds) > num_expected_iou_sampling:
                if self.num_bins >= 2:
                    iou_sampled_inds = self.sample_via_interval(
                        max_overlaps, set(iou_sampling_neg_inds),
                        num_expected_iou_sampling)
                else:
                    iou_sampled_inds = self.random_choice(
                        iou_sampling_neg_inds, num_expected_iou_sampling)
            else:
                iou_sampled_inds = np.array(
                    iou_sampling_neg_inds, dtype=np.int)
            num_expected_floor = num_expected - len(iou_sampled_inds)
            if len(floor_neg_inds) > num_expected_floor:
                sampled_floor_inds = self.random_choice(
                    floor_neg_inds, num_expected_floor)
            else:
                sampled_floor_inds = np.array(floor_neg_inds, dtype=np.int)
            sampled_inds = np.concatenate(
                (sampled_floor_inds, iou_sampled_inds))
            if len(sampled_inds) < num_expected:
                num_extra = num_expected - len(sampled_inds)
                extra_inds = np.array(list(neg_set - set(sampled_inds)))
                if len(extra_inds) > num_extra:
                    extra_inds = self.random_choice(extra_inds, num_extra)
                sampled_inds = np.concatenate((sampled_inds, extra_inds))
            sampled_inds = torch.from_numpy(sampled_inds).long().to(
                assign_result.gt_inds.device)
            return sampled_inds



import torch.nn as nn
import torch.nn.functional as F
import torch
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class FeatureLoss(nn.Module):

    """PyTorch version of `Masked Generative Distillation`
   
    Args:
        student_channels(int): Number of channels in the student's feature map.
        teacher_channels(int): Number of channels in the teacher's feature map. 
        name (str): the loss name of the layer
        alpha_mgd (float, optional): Weight of dis_loss. Defaults to 0.00002
        lambda_mgd (float, optional): masked ratio. Defaults to 0.65
    """
    def __init__(self,
                 student_channels,
                 teacher_channels,
                 name,
                 alpha_mgd=0.00002,
                 lambda_mgd=0.65,
                 ):
        super(FeatureLoss, self).__init__()
        self.alpha_mgd = alpha_mgd
        self.lambda_mgd = lambda_mgd
        self.name = name
    
        if student_channels != teacher_channels:
            self.align = nn.Conv2d(student_channels, teacher_channels, kernel_size=1, stride=1, padding=0)
        else:
            self.align = None

        self.generation = nn.Sequential(
            nn.Conv2d(teacher_channels, teacher_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True), 
            nn.Conv2d(teacher_channels, teacher_channels, kernel_size=3, padding=1))


    def forward(self,
                preds_S,
                preds_T):
        """Forward function.
        Args:
            preds_S(Tensor): Bs*C*H*W, student's feature map
            preds_T(Tensor): Bs*C*H*W, teacher's feature map
        """
        assert preds_S.shape[-2:] == preds_T.shape[-2:]

        if self.align is not None:
            preds_S = self.align(preds_S)
    
        loss = self.get_dis_loss(preds_S, preds_T)*self.alpha_mgd
            
        return loss

    def get_dis_loss(self, preds_S, preds_T):
        loss_mse = nn.MSELoss(reduction='sum')
        N, C, H, W = preds_T.shape

        device = preds_S.device
        mat = torch.rand((N,1,H,W)).to(device)
        mat = torch.where(mat>1-self.lambda_mgd, 0, 1).to(device)

        masked_fea = torch.mul(preds_S, mat)
        new_fea = self.generation(masked_fea)

        dis_loss = loss_mse(new_fea, preds_T)/N

        return dis_loss


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import astor
import sys
# from x2paddle.project_convertor.pytorch.mapper import *
import copy
import os.path as osp
from .utils import get_dep_file_path


class DepInfo:
    """
    依赖包信息。
    PT_FROM代表pytorch from信息的字符串，例如：torch；
    PD_FROM代表paddle from信息的字符串，例如：paddle；
    PT_IMPORT代表pytorch import信息系的字符串，例如：nn.functional；
    PD_IMPORT代表paddle import信息系的字符串，例如：nn.functional；
    AS代表as信息的字符串，例如：F；
    PT_DEPENDENCY代表由PT_FROM、PT_IMPORT、AS三者组成的字符串，例如：from torch import nn.functional as F。
    PD_DEPENDENCY代表由PD_FROM、PD_IMPORT、AS三者组成的字符串，例如：from paddle import nn.functional as F。
    """
    PT_FROM = None
    PD_FROM = None
    PT_IMPORT = None
    PD_IMPORT = None
    AS = None
    PT_DEPENDENCY = None
    PD_DEPENDENCY = None


class AstUpdater(ast.NodeVisitor):
    """ 更新ast树，将ast树中PyTorch相关的节点转为Paddle相关的节点。
    Args:
        py_file_path (str): python文件的绝对值路径。
        file_dependencies (dict): 当前已经统计的依赖信息，key为python文件的绝对值路径，
                                value为key文件所对应的依赖信息组成的list。
    """

    def __init__(self, py_file_path, file_dependencies):
        self.py_file_path = py_file_path
        self.root = ast.parse(open(py_file_path, "rb").read())
        self.file_dependencies = file_dependencies
        self.scopes_and_dependencies = list()  # 作用域和依赖组成的stack
        self.nodes = list()  # ast节点组成的stack
        self.no_support_apis = list()  # 不支持的API列表
        self.is_import_torch2paddle = False  # 是否添加import torch2paddle
        self.is_import_paddle = True  # 是否添加import padddle
        self.is_import_x2paddle = False  # 是否添加import x2paddle

    def _get_scope_node(self):
        """ 获取当前节点的作用域。
        """
        scope_node = None
        for i in range(len(self.scopes_and_dependencies)):
            i = -(i + 1)
            sd = self.scopes_and_dependencies[i]
            if not isinstance(sd, DepInfo) and not isinstance(sd, ast.Assign):
                scope_node = sd
                break
        return scope_node

    def _get_current_index(self, scope_node, node):
        """ 获取当前节点在其作用域中的索引序号。
        """
        current_id = 0
        for i, n in enumerate(scope_node.body):
            if node == n:
                current_id = i
                break
        return current_id

    def _get_father_node(self):
        """ 获取父节点。
        """
        return self.nodes[-2]

    def _get_complete_api(self, api_part_name):
        """ 根据部分api名字获取PyTorch的api全名。
            情况1：依赖是DepInfo，但其PD_IMPORT为None（非PyTorch的依赖），则pytorch_api为None。
            情况2：依赖是DepInfo，且DepInfo的部分PyTorch属性以“torch”开头，则pytorch_api为完整api。
            情况3：依赖是ast.Assign节点，则pytorch_api为None。
        """
        pytorch_api = None
        dep_info = None
        if api_part_name is None:
            return pytorch_api, dep_info
        for i in range(len(self.scopes_and_dependencies)):
            i = -(i + 1)
            dep_info = self.scopes_and_dependencies[i]
            if isinstance(dep_info, DepInfo):
                if dep_info.PT_IMPORT is None:
                    continue
                if (dep_info.PT_FROM is not None and "torch" in dep_info.PT_FROM) or \
                        (dep_info.PT_IMPORT is not None and "torch" in dep_info.PT_IMPORT):
                    replace_str = None
                    if dep_info.AS is not None and api_part_name.startswith(
                            dep_info.AS + "."):
                        replace_str = dep_info.AS
                    elif dep_info.AS is None and api_part_name.startswith(
                            dep_info.PT_IMPORT):
                        replace_str = dep_info.PT_IMPORT
                    if replace_str is not None:
                        pytorch_api = api_part_name.replace(
                            replace_str, dep_info.PT_DEPENDENCY, 1)
                        if "torch2paddle" in pytorch_api:
                            # 说明当前节点是插入的已经替换过的node
                            pytorch_api = None
                        break
            elif isinstance(dep_info, ast.Assign):
                is_customized = False
                for s in astor.to_source(dep_info.targets[0]).split(","):
                    if api_part_name.split(".")[0] == s.strip():
                        is_customized = True
                        break
                if is_customized:
                    break
        return pytorch_api, dep_info

    def _rename(self, name, dep_info, pytorch_api, paddle_api):
        """ 对函数名进行重命名。
            例如：将nn.Conv2d替换为nn.Conv2D。
        """
        pytorch_api_seg = pytorch_api.split(dep_info.PT_IMPORT)
        if ".models." in paddle_api:
            self.is_import_x2paddle = True
        if paddle_api.startswith(dep_info.PD_IMPORT + ".") or \
                paddle_api.endswith("." + dep_info.PD_IMPORT) or  \
                "." + dep_info.PD_IMPORT + "." in paddle_api:
            paddle_api_seg = paddle_api.split(dep_info.PD_IMPORT)
            if dep_info.AS is None:
                name = name.replace(dep_info.PT_IMPORT + pytorch_api_seg[-1],
                                    dep_info.PD_IMPORT + paddle_api_seg[-1])
            else:
                name = name.replace(pytorch_api_seg[-1], paddle_api_seg[-1])
        elif "torch2paddle." in paddle_api:
            name = "torch2paddle." + paddle_api.split("torch2paddle.")[-1]
            self.is_import_torch2paddle = True
        else:
            name = paddle_api
        return name

    def run(self):
        self.scopes_and_dependencies.append(self.root)
        self.visit(self.root)
        for i, node in enumerate(self.root.body):
            if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
                if self.is_import_torch2paddle:
                    self.root.body.insert(
                        i,
                        ast.parse("from x2paddle import torch2paddle").body[0])
                if self.is_import_x2paddle:
                    self.root.body.insert(i,
                                          ast.parse("import x2paddle").body[0])
                if self.is_import_paddle:
                    self.root.body.insert(i, ast.parse("import paddle").body[0])
                break

    def visit(self, node):
        self.nodes.append(node)
        out = super(AstUpdater, self).visit(node)
        self.nodes.pop()
        if out is not None:
            return out
        else:
            # 出现字符串或者if等节点需要返回字符串
            try:
                return astor.to_source(node)
            except Exception:
                return None

    def visit_ImportFrom(self, node):
        """ 1. 遍历子节点。
            2. 将当前from依赖中的多个import拆分成多个import。
                例如：from torch import nn, utils 这个node
                     拆分为：node1：from torch import nn
                            node2：from torch import utils
                拆分原因：
                    在paddle中父依赖包可能不一致。
        """
        scope_node = self._get_scope_node()
        current_id = self._get_current_index(scope_node, node)
        scope_node.body.pop(current_id)
        son_nodes = node.names
        for i, son_node in enumerate(son_nodes):
            copy_node = copy.deepcopy(node)
            copy_node.names = [son_node]
            if i == 0:
                is_remove = self.visit_alias(son_node, copy_node, node.module,
                                             node.level)
                if not is_remove:
                    scope_node.body.insert(current_id, copy_node)
            else:
                scope_node.body.insert(current_id + i, copy_node)

    def visit_Import(self, node):
        """ 遍历子节点。
        """
        son_nodes = getattr(node, "names")
        for son_node in son_nodes:
            self.visit_alias(son_node, node)

    def visit_alias(self,
                    node,
                    father_node=None,
                    from_name=None,
                    from_level=None):
        """ 构建DepInfo并将其放入scopes_and_dependencies。
            如果import字符串为“*”，获取依赖包所在文件的依赖信息并转换为DepInfo加入当前的scopes_and_dependencies；
            反之，直接在scopes_and_dependencies中加入DepInfo。
        """
        is_remove = False
        dep_info = DepInfo()
        dep_info.PT_FROM = from_name
        dep_info.PT_IMPORT = getattr(node, "name")
        dep_info.AS = getattr(node, "asname", None)
        if dep_info.PT_IMPORT == "*":
            import_file_path = get_dep_file_path(self.py_file_path, from_level,
                                                 from_name)
            pytorch_dependencies = self.file_dependencies[import_file_path]
            for pytorch_dep_info in pytorch_dependencies:
                current_dep_info = DepInfo()
                if not isinstance(pytorch_dep_info, str):
                    current_dep_info.PT_FROM = pytorch_dep_info.FROM
                    current_dep_info.PT_IMPORT = pytorch_dep_info.IMPORT
                    current_dep_info.AS = pytorch_dep_info.AS
                    current_dep_info.PT_DEPENDENCY = pytorch_dep_info.DEPENDENCY
                    if "torch" in current_dep_info.PT_DEPENDENCY:
                        if current_dep_info.PT_DEPENDENCY in API_MAPPER:
                            current_dep_info.PD_DEPENDENCY = \
                                    API_MAPPER[current_dep_info.PT_DEPENDENCY][0]
                            if current_dep_info.PT_DEPENDENCY == "torch" and \
                                    isinstance(self._get_scope_node(), ast.Module):
                                self.is_import_paddle = False
                            if current_dep_info.PT_FROM is not None:
                                seg = current_dep_info.PD_DEPENDENCY.split(".")
                                current_dep_info.PD_IMPORT = seg[-1]
                                current_dep_info.PD_FROM = \
                                        current_dep_info.PD_DEPENDENCY.replace("." + seg[-1], "")
                            else:
                                current_dep_info.PD_IMPORT = \
                                current_dep_info.PD_DEPENDENCY
                        elif current_dep_info.PT_DEPENDENCY in REMOVE_API:
                            scope_node = self._get_scope_node()
                            for i, n in enumerate(scope_node.body):
                                if father_node == n:
                                    scope_node.body[i] = ast.parse("\n").body
                            is_remove = True
                        else:
                            self.no_support_apis.append(
                                current_dep_info.PT_DEPENDENCY)
                else:
                    current_dep_info.PD_DEPENDENCY = pytorch_dep_info
                self.scopes_and_dependencies.append(current_dep_info)
            return is_remove
        dependency_str_list = list()
        if dep_info.PT_FROM is None and from_level is not None:
            dependency_str_list.append("." * from_level)
        elif dep_info.PT_FROM is not None:
            dependency_str_list.append(dep_info.PT_FROM)
        dependency_str_list.append(dep_info.PT_IMPORT)
        dep_info.PT_DEPENDENCY = ".".join(dependency_str_list)
        if dep_info.PT_DEPENDENCY.startswith("torch"):
            if dep_info.PT_DEPENDENCY in API_MAPPER:
                dep_info.PD_DEPENDENCY = API_MAPPER[dep_info.PT_DEPENDENCY][0]
                if dep_info.PT_DEPENDENCY == "torch":
                    self.is_import_paddle = False
                if dep_info.PT_FROM is not None:
                    seg = dep_info.PD_DEPENDENCY.split(".")
                    setattr(node, "name", seg[-1])
                    setattr(father_node, "module",
                            dep_info.PD_DEPENDENCY.replace("." + seg[-1], ""))
                    dep_info.PD_IMPORT = seg[-1]
                    dep_info.PD_FROM = dep_info.PD_DEPENDENCY.replace(
                        "." + seg[-1], "")
                else:
                    setattr(node, "name", dep_info.PD_DEPENDENCY)
                    dep_info.PD_IMPORT = dep_info.PD_DEPENDENCY
            elif dep_info.PT_DEPENDENCY in REMOVE_API:
                scope_node = self._get_scope_node()
                for i, n in enumerate(scope_node.body):
                    if father_node == n:
                        scope_node.body[i] = ast.parse("\n").body
                is_remove = True
            elif dep_info.PT_DEPENDENCY.startswith("torch"):
                self.no_support_apis.append(dep_info.PT_DEPENDENCY)
        else:
            dep_info.PD_DEPENDENCY = dep_info.PT_DEPENDENCY
        self.scopes_and_dependencies.append(dep_info)
        return is_remove

    def visit_Name(self, node):
        """ 获取字符串名字。
        """
        pytorch_api, dep_info = self._get_complete_api(getattr(node, "id"))
        father_node = self._get_father_node()
        if pytorch_api in API_MAPPER:
            paddle_api = API_MAPPER[pytorch_api][0]
            if isinstance(father_node, ast.Call) and getattr(
                    father_node.func, "id", None) in ("getattr", "setattr",
                                                      "hasattr"):
                paddle_api = self._rename(paddle_api, dep_info, pytorch_api,
                                          paddle_api)
                for i, arg_node in enumerate(father_node.args):
                    if astor.to_source(arg_node).strip() == getattr(node, "id"):
                        father_node.args[i] = ast.parse(paddle_api).body[
                            0].value
        return getattr(node, "id")

    def visit_Attribute(self, node):
        """ 对属性字符串满足以下4种情况时进行替换：
            情况1 —— Class A(nn.Module)：将nn.Module替换为nn.Layer；
            情况2 —— a = (1, 2, nn.Module)：将nn.Module替换为nn.Layer；
            情况3 —— def a() -> torch.Tensor：将torch.Tensor替换为paddle.Tensor；
            情况4 —— def a(x: torch.Tensor)：将torch.Tensor替换为paddle.Tensor；
            情况5 —— isinstance(a, nn.Module)：将nn.Module替换为nn.Layer;
            情况6 —— torch.float32：将torch.float32替换为"float32"；
        """
        value_node = node.value
        attr = node.attr
        name = self.visit(value_node)
        attr_str = name + "." + attr
        pytorch_api, dep_info = self._get_complete_api(attr_str)
        father_node = self._get_father_node()
        if pytorch_api in API_MAPPER:
            paddle_api = API_MAPPER[pytorch_api][0]
            if isinstance(father_node, ast.ClassDef):
                attr_str = self._rename(attr_str, dep_info, pytorch_api,
                                        paddle_api)
                if node in father_node.bases:
                    father_node.bases[0] = ast.parse(attr_str).body[0].value
                return attr_str
            elif isinstance(father_node, ast.arguments):
                attr_str = self._rename(attr_str, dep_info, pytorch_api,
                                        paddle_api)
                for i, default_n in enumerate(father_node.defaults):
                    if default_n == node:
                        father_node.defaults[i] = ast.parse(attr_str).body[
                            0].value
                return attr_str
            elif isinstance(father_node, ast.Tuple):
                paddle_api = self._rename(paddle_api, dep_info, pytorch_api,
                                          paddle_api)
                for i, elts_node in enumerate(father_node.elts):
                    if astor.to_source(elts_node).strip() == attr_str:
                        father_node.elts[i] = ast.parse(paddle_api).body[
                            0].value
                return paddle_api
            elif isinstance(father_node, ast.FunctionDef):
                paddle_api = self._rename(paddle_api, dep_info, pytorch_api,
                                          paddle_api)
                father_node.returns = ast.parse(paddle_api).body[0].value
                return paddle_api
            elif isinstance(father_node, ast.arg):
                attr_str = self._rename(attr_str, dep_info, pytorch_api,
                                        paddle_api)
                father_node.annotation = ast.parse(attr_str).body[0].value
                return attr_str
            elif isinstance(father_node, ast.Call) and getattr(
                    father_node.func, "id", None) == "isinstance":
                paddle_api = self._rename(paddle_api, dep_info, pytorch_api,
                                          paddle_api)
                for i, arg_node in enumerate(father_node.args):
                    if astor.to_source(arg_node).strip() == attr_str:
                        father_node.args[i] = ast.parse(paddle_api).body[
                            0].value
                return paddle_api
            elif not isinstance(father_node, ast.Call):
                # 对torch.float32的处理
                for k, v in father_node.__dict__.items():
                    if v == node:
                        father_node.k = ast.parse(paddle_api).body[0].value
                        break
                return attr_str
        elif pytorch_api in REMOVE_API:
            if isinstance(
                    father_node,
                (ast.Assign, ast.If, ast.FunctionDef, ast.ClassDef, ast.Call)):
                scope_node = self._get_scope_node()
                for i, n in enumerate(scope_node.body):
                    if father_node == n:
                        scope_node.body.pop(i)
                        return None
            elif isinstance(father_node, ast.BoolOp):
                for i, n in enumerate(father_node.values):
                    if node == n:
                        father_node.values[i] = ast.parse("False").body[0].value
                        return None
        else:
            if isinstance(pytorch_api, str) and pytorch_api.startswith(
                    "torch"
            ) and "(" not in pytorch_api and "[" not in pytorch_api:
                if not isinstance(father_node, ast.Attribute):
                    self.no_support_apis.append(pytorch_api)
        return attr_str

    def visit_Num(self, node):
        """ 返回数值。
        """
        return getattr(node, "n")

    def visit_keyword(self, node):
        """ 返回键值对。
            【注意】当value是API_MAPPER中的key时，需要替换为API_MAPPER中对应的Paddle API。
        """
        key = getattr(node, "arg")
        value_node = getattr(node, "value")
        value = self.visit(value_node)
        if value in API_MAPPER:
            value = API_MAPPER[value][0]
        elif isinstance(value, str) and value.startswith(
                "torch") and "(" not in value and "[" not in value:
            self.no_support_apis.append(value)
        return {key: value}

    def visit_Tuple(self, node):
        """ 返回tuple。
        """
        elts_nodes = getattr(node, "elts")
        elts = list()
        for elts_node in elts_nodes:
            elt = self.visit(elts_node)
            elts.append(elt if isinstance(elt, str) else str(elt))
        elts = "({})".format(", ".join(elts))
        return elts

    def visit_Assign(self, node):
        """ 1. 将Assign节点加入scopes_and_dependencies；
            2. 遍历Assign节点的子节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)

    def visit_Call(self, node):
        """ 1. 获取原始函数名并更新为新的函数名。
            2. 获取args和kwargs。
            3. 根据API_MAPPER映射需要更新的操作，对参数进行处理。
            4. 如果有前缀代码和后缀代码，则需要添加相应节点。
        """
        # 获取函数名
        func_node = node.func
        if isinstance(func_node, ast.Attribute) and isinstance(func_node.value,
                                                               ast.Call):
            func_name = None
        else:
            func_name = self.visit(func_node)
        pytorch_api, dep_info = self._get_complete_api(func_name)
        if pytorch_api is None:
            self.generic_visit(node)
            return
        if pytorch_api not in API_MAPPER:
            if pytorch_api.startswith(
                    "torch"
            ) and "[" not in pytorch_api and "(" not in pytorch_api:
                self.no_support_apis.append(pytorch_api)
            return
        paddle_api = API_MAPPER[pytorch_api][0]
        func_name = self._rename(func_name, dep_info, pytorch_api, paddle_api)
        setattr(node, "func", ast.parse(func_name).body[0].value)

        # 获取args
        args_nodes = getattr(node, "args")
        args_list = list()
        for args_node in args_nodes:
            args_list.append(self.visit(args_node))

        # 获取keywords
        keywords_nodes = getattr(node, "keywords")
        kw_dict = dict()
        for keywords_node in keywords_nodes:
            if list(self.visit(keywords_node).keys())[0] is None:
                args_list.append("**{}".format(
                    list(self.visit(keywords_node).values())[0]))
            else:
                kw_dict.update(self.visit(keywords_node))

        if API_MAPPER[pytorch_api][1] is None:
            return

        target_name = None
        father_node = self._get_father_node()
        if isinstance(father_node, ast.Assign):
            target_node = father_node.targets[0]
            target_name = self.visit(target_node)
        mapper = API_MAPPER[pytorch_api][1](func_name, pytorch_api, args_list,
                                            kw_dict, target_name)
        prefix_insert_codes, new_code, suffix_insert_codes = mapper.run()
        if mapper.func_name.startswith("x2paddle."):
            self.is_import_x2paddle = True
        scope_node = self._get_scope_node()
        if isinstance(ast.parse(new_code).body[0], ast.Assign):
            node_index = self._get_current_index(scope_node, node)
            scope_node.body[node_index] = ast.parse(
                new_code.replace("\n", "")).body[0]
        else:
            new_call_node = ast.parse(new_code).body[0].value
            setattr(node, "func", new_call_node.func)  # 修改了fun_name
            setattr(node, "args", new_call_node.args)
            setattr(node, "keywords", new_call_node.keywords)
        for i, n in enumerate(scope_node.body):
            if father_node == n:
                for code in prefix_insert_codes:
                    scope_node.body.insert(
                        i, ast.parse(code.replace("\n", "")).body[0])
                    i += 1
                break
        for i, n in enumerate(scope_node.body):
            if father_node == n:
                j = i + 1
                for code in suffix_insert_codes:
                    scope_node.body.insert(
                        j, ast.parse(code.replace("\n", "")).body[0])
                    j += 1
                break

    def visit_Subscript(self, node):
        value_node = node.value
        value_name = self.visit(value_node)
        pytorch_api, dep_info = self._get_complete_api(value_name)
        if pytorch_api in API_MAPPER:
            paddle_api = API_MAPPER[pytorch_api][0]
            value_name = self._rename(value_name, dep_info, pytorch_api,
                                      paddle_api)
            node.value = ast.parse(value_name).body[0]
        else:
            if isinstance(pytorch_api, str) and pytorch_api.startswith(
                    "torch") and "(" not in pytorch_api:
                self.no_support_apis.append(pytorch_api)
        self.visit(node.slice)
        self.visit(node.ctx)

    def visit_FunctionDef(self, node):
        """ 1. 将FunctionDef节点加入scopes_and_dependencies；
            2. 遍历FunctionDef节点的子节点；
            3. 去除scopes_and_dependencies中FunctionDef节点以及之后的节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)
        last_node = self.scopes_and_dependencies.pop(-1)
        while not isinstance(last_node, ast.FunctionDef):
            last_node = self.scopes_and_dependencies.pop(-1)

    def visit_ClassDef(self, node):
        """ 1. 将ClassDef节点加入scopes_and_dependencies；
            2. 遍历ClassDef节点的子节点；
            3. 去除scopes_and_dependencies中ClassDef节点以及之后的节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)
        last_node = self.scopes_and_dependencies.pop(-1)
        while not isinstance(last_node, ast.ClassDef):
            last_node = self.scopes_and_dependencies.pop(-1)

    def visit_If(self, node):
        """ 1. 将If节点加入scopes_and_dependencies；
            2. 遍历If节点的子节点；
            3. 去除scopes_and_dependencies中If节点以及之后的节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)
        last_node = self.scopes_and_dependencies.pop(-1)
        while not isinstance(last_node, ast.If):
            last_node = self.scopes_and_dependencies.pop(-1)

    def visit_While(self, node):
        """ 1. 将While节点加入scopes_and_dependencies；
            2. 遍历Try节点的子节点；
            3. 去除scopes_and_dependencies中Try节点以及之后的节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)
        last_node = self.scopes_and_dependencies.pop(-1)
        while not isinstance(last_node, ast.While):
            last_node = self.scopes_and_dependencies.pop(-1)

    def visit_Try(self, node):
        """ 1. 将Try节点加入scopes_and_dependencies；
            2. 遍历Try节点的子节点；
            3. 去除scopes_and_dependencies中Try节点以及之后的节点。
        """
        self.scopes_and_dependencies.append(node)
        self.generic_visit(node)
        last_node = self.scopes_and_dependencies.pop(-1)
        while not isinstance(last_node, ast.Try):
            last_node = self.scopes_and_dependencies.pop(-1)

    def visit_ExtSlice(self, node):
        """ 将Index节点替换替换为Num。
        """
        dim_nodes = node.dims
        for i, dim_node in enumerate(dim_nodes):
            if isinstance(dim_node, ast.Index):
                dim_nodes[i] = dim_node.value
            else:
                self.visit(dim_node)

    def visit_Str(self, node):
        """ 修改模型参数的后缀名。
        """
        setattr(node, "s",
                node.s.replace(".pth", ".pdiparams").replace(
                    ".pt", ".pdiparams").replace(".ckpt", ".pdiparams"))


def update(py_file_path, file_dependencies):
    updater = AstUpdater(py_file_path, file_dependencies)
    updater.run()
    if len(updater.no_support_apis) > 0:
        print("Can not convert the file {}.".format(py_file_path))
        print("The unsupported packages or operators are: [{}].".format(
            ", ".join(set(updater.no_support_apis))))
        return None
    else:
        return updater.root



"""Backend Manager."""
import importlib
from itertools import chain
from .common import MobulaTensor

DTYPE_TO_GLUE = dict()  # dtype -> glue_mod
GLUE_NAME_TO_GLUE = dict()  # glue_name -> glue_mod
PKG_NAME_TO_GLUE_ARGS = dict()  # package_name -> (glue_name, types_name)


def _register_glue_real(glue_name, types_name):
    global DTYPE_TO_GLUE, GLUE_NAME_TO_GLUE
    if not isinstance(types_name, list):
        types_name = [types_name]
    glue = None
    try:
        glue = importlib.import_module('.' + glue_name + '_glue', __package__)
    except ImportError:
        pass
    if glue is not None:
        for tname in types_name:
            tname_sp = tname.split('.')
            try:
                module = importlib.import_module(tname_sp[0])
                for sub_name in tname_sp[1:]:
                    module = getattr(module, sub_name, None)
                    if module is None:
                        raise ImportError
                # create generators cache
                glue.gen_cache = dict()
                DTYPE_TO_GLUE[module] = glue
            except ImportError:
                pass
            # assert hasattr(glue.OpGen, '__call__')
            # assert hasattr(glue.OpGen, 'register')
            if hasattr(glue, 'Tensor'):
                value = glue.Tensor
                assert isinstance(value, type) and issubclass(
                    value, MobulaTensor)
            else:
                # register glue tensor
                tensors = []
                for key, value in glue.__dict__.items():
                    if isinstance(value, type) and issubclass(value, MobulaTensor):
                        if key.lower() == glue_name + 'tensor':
                            tensors.append(key)
                assert len(
                    tensors) == 1, 'Only one MobulaTensor is allowed in each glue module'
                glue.Tensor = getattr(glue, tensors[0])

            GLUE_NAME_TO_GLUE[glue_name] = glue


def register_glue(glue_name, type_names):
    """Register a glue module.

    Parameters
    ----------
    glue_name: str
        The name of glue module.
    type_names: list of str
        The list of inputs' class names.
    """
    global PKG_NAME_TO_GLUE_ARGS
    assert type_names, ValueError('type_names should be not empty')
    pkg_names = [cls_name.split('.')[0] for cls_name in type_names]
    pkg_name = pkg_names[0]
    assert all(map(lambda x: x == pkg_name, pkg_names)), TypeError(
        'The name of package should be consistent in `types_name`: {}'.format(type_names))
    PKG_NAME_TO_GLUE_ARGS[pkg_name] = (glue_name, type_names)


# register glue modules.
register_glue('mxnet', ['mxnet.nd.NDArray',
                        'mxnet.sym.Symbol', 'mxnet.numpy.ndarray'])
register_glue('numpy', ['numpy.ndarray'])
register_glue('torch', ['torch.Tensor'])
register_glue('cupy', ['cupy.core.core.ndarray'])


def get_var_type_glue(vtype):
    """Get glue module from variable's type.

    Parameters
    ----------
    vtype: data type

    Returns
    -------
    Glue Module if glue exists, otherwise None.
    """
    global DTYPE_TO_GLUE, PKG_NAME_TO_GLUE_ARGS
    glue_mod = DTYPE_TO_GLUE.get(vtype, None)
    if glue_mod is not None:
        return glue_mod
    pkg_name = vtype.__module__.split('.')[0]
    if pkg_name not in PKG_NAME_TO_GLUE_ARGS:
        return None
    # try to register glue_mod
    _register_glue_real(*PKG_NAME_TO_GLUE_ARGS[pkg_name])
    return DTYPE_TO_GLUE[vtype]


def get_var_glue(var):
    """Get glue module from variable.

    Parameters
    ----------
    var: variable

    Returns
    -------
    Glue Module if glue exists, otherwise None.
    """

    return get_var_type_glue(type(var))


def get_args_glue(*args, **kwargs):
    """Get glue module from args and kwargs.

    Parameters
    ----------
    *args
    **kwargs

    Returns
    -------
    Glue Module if glue exists, otherwise None.
    """
    glue_mods = map(get_var_glue, chain(args, kwargs.values()))
    glue_mods = list(filter(lambda x: x is not None, glue_mods))
    if glue_mods:
        glue_mod = glue_mods[0]
        assert all(map(lambda x: x == glue_mod, glue_mods)),\
            TypeError(
                'Support only 1 backend in a call, now: {}'.format(glue_mods))
        return glue_mod
    return None


def op_gen(glue_mod, op, name):
    """ Get operator generator of glue module.

    Parameters
    ----------
    glue_mod: Glue Module
    op: object
        The object of custom operator.
    name: str
        The name of custom operator.

    Returns
    -------
    The operator generator of glue module.
    """
    if name not in glue_mod.gen_cache:
        glue_mod.gen_cache[name] = glue_mod.OpGen(op=op, name=name)
    return glue_mod.gen_cache[name]


def get_glue_modules():
    return GLUE_NAME_TO_GLUE.values()



# Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# function:
#   transform samples in 'source' using 'worker'

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import six
if six.PY3:
    from queue import Empty
else:
    from Queue import Empty

import uuid
import logging
import signal
import threading
import traceback
from antgo.utils import logger
main_pid = os.getpid()
worker_set = set()


class EndSignal(object):
    """ signal used to notify worker to exit
    """

    def __init__(self, id, errno=0, errmsg=''):
        self.id = id
        self.errno = errno
        self.errmsg = errmsg


class ParallelMap(object):
    """
    Transform samples to mapped samples which is similar to 
    'basic.MappedDataset', but multiple workers (threads or processes) 
    will be used

    Notes:
        this class is not thread-safe
    """

    def __init__(self,
                 source,
                 worker,
                 worker_num,
                 bufsize=100,
                 use_process=False,
                 memsize='30G'):
        self._worker_num = worker_num
        self._bufsize = bufsize
        self._use_process = use_process
        if self._use_process and sys.platform == "win32":
            logger.debug("Use multi-thread reader instead of "
                         "multi-process reader on Windows.")
            self._use_process = False
        if self._use_process and type(memsize) is str:
            assert memsize[-1].lower() in ['g', 'm'], \
                "invalid param for memsize[%s], should be " \
                "ended with 'G' or 'g' or 'M' or 'm'" % (memsize)
            power = 3 if memsize[-1].lower() == 'g' else 2
            self._memsize = int(memsize[:-1]) * (1024**power)
        self._started = False
        self._source = source
        self._worker = worker
        self._exit = False
        self._setup()
        self._souce_drained = False

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def _setup(self):
        """setup input/output queues and workers """
        use_process = self._use_process

        bufsize = self._bufsize
        if use_process:
            from .shared_queue import SharedQueue as Queue
            from multiprocessing import Process as Worker
            from multiprocessing import Event
            memsize = self._memsize
            self._inq = Queue(bufsize, memsize=memsize)
            self._outq = Queue(bufsize, memsize=memsize)
        else:
            if six.PY3:
                from queue import Queue
            else:
                from Queue import Queue
            from threading import Thread as Worker
            from threading import Event
            self._inq = Queue(bufsize)
            self._outq = Queue(bufsize)

        consumer_num = self._worker_num
        id = str(uuid.uuid4())[-3:]
        self._producer = threading.Thread(
            target=self._produce,
            args=('producer-' + id, self._source, self._inq))
        self._producer.daemon = True

        self._consumers = []
        self._consumer_endsig = {}
        global worker_set
        for i in range(consumer_num):
            consumer_id = 'consumer-' + id + '-' + str(i)
            p = Worker(
                target=self._consume,
                args=(consumer_id, self._inq, self._outq, self._worker))
            self._consumers.append(p)
            p.daemon = True
            setattr(p, 'id', consumer_id)
            if use_process:
                worker_set.add(p)

        self._epoch = -1
        self._feeding_ev = Event()
        self._produced = 0  # produced sample in self._produce
        self._consumed = 0  # consumed sample in self.next

    def _produce(self, id, source, inq):
        """Fetch data from source and feed it to 'inq' queue"""
        endsig = EndSignal(id)
        while True:
            self._feeding_ev.wait()
            if self._exit:
                break
            try:
                s = source.next()
                inq.put(s)
                self._produced += 1
            except StopIteration:
                self._souce_drained = True
                self._feeding_ev.clear()
                self._feeding_ev.wait()
            except Exception as e:
                endsig.errno = -1
                endsig.errmsg = "producer[{}] failed with error: {}" \
                    .format(id, str(e))
                inq.put(endsig)
                break

    def _consume(self, id, inq, outq, worker):
        """Fetch data from 'inq', process it and put result to 'outq'"""
        if self._use_process:
            # handle SIGTERM signal to exit to prevent print stack frame
            signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())

        endsig = EndSignal(id)
        while True:
            sample = inq.get()
            if isinstance(sample, EndSignal):
                endsig.errno = sample.errno
                endsig.errmsg = "consumer[{}] exits for reason[{}]" \
                    .format(id, sample.errmsg)
                outq.put(endsig)
                break

            try:
                result = worker(sample)
                outq.put(result)
            except Exception as e:
                endsig.errno = -2
                endsig.errmsg = "consumer[{}] failed to map with error:[{}]" \
                    .format(id, str(e))
                outq.put(endsig)
                break

    def drained(self):
        assert self._epoch >= 0, "first epoch has not started yet"
        return self._source.drained() and self._produced == self._consumed

    def stop(self):
        """ notify to exit
        """
        self._exit = True
        self._feeding_ev.set()
        for _ in range(len(self._consumers)):
            self._inq.put(EndSignal(0, "notify consumers to exit"))

    def _consumer_healthy(self):
        abnormal_num = 0
        for w in self._consumers:
            if not w.is_alive() and w.id not in self._consumer_endsig:
                abnormal_num += 1
                if self._use_process:
                    errmsg = "consumer[{}] exit abnormally with exitcode[{}]" \
                                .format(w.pid, w.exitcode)
                else:
                    errmsg = "consumer[{}] exit abnormally".format(w.ident)

                logger.warn(errmsg)

        if abnormal_num > 0:
            logger.warn("{} consumers have exited abnormally!!!" \
                .format(abnormal_num))

        return abnormal_num == 0

    def next(self):
        """ get next transformed sample
        """
        if self._epoch < 0:
            self.reset()

        if self.drained():
            raise StopIteration()

        while not self._exit:
            try:
                sample = self._outq.get(timeout=3)
            except Empty as e:
                if not self._consumer_healthy():
                    raise StopIteration()
                else:
                    continue

            if isinstance(sample, EndSignal):
                self._consumer_endsig[sample.id] = sample
                logger.warn("recv endsignal from outq with errmsg[{}]" \
                    .format(sample.errmsg))

                if len(self._consumer_endsig.keys()) < len(self._consumers):
                    self._inq.put(sample)
                else:
                    self._exit = True
                    raise StopIteration("all consumers exited, no more samples")
            else:
                self._consumed += 1
                return sample

        raise StopIteration()

    def reset(self):
        """ reset for a new epoch of samples
        """
        assert not self._exit, "cannot reset for already stopped dataset"

        if self._epoch < 0:
            self._epoch = 0
            for w in self._consumers:
                w.start()
            self._producer.start()
        else:
            assert self._consumer_healthy(), "cannot start another pass of data" \
                " for some consumers exited abnormally before!!!"

            if not self.drained():
                logger.warn("reset before epoch[{}] finishes".format(
                    self._epoch))
                self._produced = self._produced - self._consumed
            else:
                self._produced = 0

            self._epoch += 1

        assert len(self._consumer_endsig.keys()) == 0, "some consumers already exited," \
            + " cannot start another epoch"

        self._source.reset()
        self._souce_drained = False
        self._consumed = 0
        self._feeding_ev.set()


# FIXME: fix me if you have better impliment
# handle terminate reader process, do not print stack frame
signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())


# FIXME(dkp): KeyboardInterrupt should be handled inside ParallelMap
# and do such as: 1. exit workers 2. close queues 3. release shared
# memory, HACK KeyboardInterrupt with global signal.SIGINT handler
# here, should be refined later
def _term_workers(sig_num, frame):
    global worker_set, main_pid
    # only do subporcess killing in main process
    if os.getpid() != main_pid:
        return

    logger.info("KeyboardInterrupt: main proc {} exit, kill subprocess {}" \
                .format(os.getpid(), [w.pid for w in worker_set]))
    for w in worker_set:
        if w.pid is not None:
            os.kill(w.pid, signal.SIGINT)
    sys.exit()


signal.signal(signal.SIGINT, _term_workers)



from inspect import trace
import sys
sys.path.append('/opt/tiger/handdetJ/antgo')
import os
import argparse
from antgo.framework.helper.tools import args
from antgo.framework.helper.utils.config import Config
import pprint
import copy
import traceback


def main(exp_args):
    # 根据传入的参数进行实验规划
    exp_schedule_config = Config.fromfile(exp_args.config)
    # 运行脚本
    exp_launch_script = exp_schedule_config.exp.launch_script

    # 运行参数
    script_args = exp_schedule_config.exp.args

    # 实验规划参数
    exp_num = len(exp_schedule_config.exp.schedule)

    pp = pprint.PrettyPrinter(indent=4)
    for exp_i in range(exp_num):
        # 打印实验进度
        print(f'ANTGO Exp Progress {exp_i+1}/{exp_num}')

        # 打印实验参数设置
        exp_config_info = copy.deepcopy(exp_schedule_config.exp.schedule[exp_i])
        exp_config_info.update({
            'launch_script': f'{exp_launch_script} {script_args}'
        })
        pp.pprint(exp_config_info)

        exp_args = ''
        for exp_arg_k, exp_arg_v in exp_schedule_config.exp.schedule[exp_i].items():
            exp_args += f'{exp_arg_k} {exp_arg_v}'

        exp_launch_cmd = f'{exp_launch_script} {script_args} {exp_args}'

        # 提示实验开始
        try:
            os.system(exp_launch_cmd)
        except Exception as e:
            traceback.print_exc()

    # 提示实验结束
    print('ANTGO Exp Finish')

if __name__ == "__main__":
    exp_args = args.nn_args()
    main(exp_args)


from antgo.framework.helper.models.nas.tutorial import AccuracyPredictor, FLOPsTable, LatencyTable, EvolutionFinder
from antgo.framework.helper.models.nas.tutorial import evaluate_ofa_subnet, evaluate_ofa_specialized
from antgo.framework.helper.models.nas.model_zoo import ofa_net
import time
from torchvision import transforms, datasets
import math
import torch
import os

def build_val_transform(size):
    return transforms.Compose([
        transforms.Resize(int(math.ceil(size / 0.875))),
        transforms.CenterCrop(size),
        transforms.ToTensor(),
        transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225]
        ),
    ])

def run():
    imagenet_data_path = '/mnt/bd/jiantrainer/dataset/imagenet'
    data_loader = torch.utils.data.DataLoader(
        datasets.ImageFolder(
            root=os.path.join(imagenet_data_path, 'val'),
            transform=build_val_transform(224)
        ),
        batch_size=250,  # test batch size
        shuffle=True,
        num_workers=16,  # number of workers for the data loader
        pin_memory=True,
        drop_last=False,
    )
    print('The ImageNet dataloader is ready.')

    ofa_network = ofa_net('ofa_mbv3_d234_e346_k357_w1.2', pretrained=True)
    print('The OFA Network is ready.')    


    target_hardware = 'note10'
    latency_table = LatencyTable(device=target_hardware)
    print('The Latency lookup table on %s is ready!' % target_hardware)

    # accuracy predictor
    accuracy_predictor = AccuracyPredictor(
        pretrained=True,
        device='cuda:0'
    )

    print('The accuracy predictor is ready!')
    print(accuracy_predictor.model)
        
    latency_constraint = 25  # ms, suggested range [15, 33] ms
    P = 100  # The size of population in each generation
    N = 500  # How many generations of population to be searched
    r = 0.25  # The ratio of networks that are used as parents for next generation
    params = {
        'constraint_type': target_hardware, # Let's do FLOPs-constrained search
        'efficiency_constraint': latency_constraint,
        'mutate_prob': 0.1, # The probability of mutation in evolutionary search
        'mutation_ratio': 0.5, # The ratio of networks that are generated through mutation in generation n >= 2.
        'efficiency_predictor': latency_table, # To use a predefined efficiency predictor.
        'accuracy_predictor': accuracy_predictor, # To use a predefined accuracy_predictor predictor.
        'population_size': P,
        'max_time_budget': N,
        'parent_ratio': r,
    }

    # build the evolution finder
    finder = EvolutionFinder(**params)

    # start searching
    result_lis = []
    st = time.time()
    best_valids, best_info, others_info = finder.run_evolution_search()
    result_lis.append(best_info)
    result_lis.extend(others_info)
    ed = time.time()
    print('Found best architecture on %s with latency <= %.2f ms in %.2f seconds! '
        'It achieves %.2f%s predicted accuracy with %.2f ms latency on %s.' %
        (target_hardware, latency_constraint, ed-st, best_info[0] * 100, '%', best_info[-1], target_hardware))

    # visualize the architecture of the searched sub-net
    _, net_config, latency = best_info
    ofa_network.set_active_subnet(ks=net_config['ks'], d=net_config['d'], e=net_config['e'])
    print('Architecture of the searched sub-net:')
    print(ofa_network.module_str)

    top1s = []
    latency_list = []
    for result in result_lis:
        acc, net_config, latency = result
        print('Evaluating the sub-network with latency = %.1f ms acc = %.6f on %s' % (latency, acc, target_hardware))
        top1 = evaluate_ofa_subnet(
            ofa_network,
            imagenet_data_path,
            net_config,
            data_loader,
            batch_size=250,
            device='cuda:0')
        print(f'top 1 {top1}')
        top1s.append(top1)
        latency_list.append(latency)


# -*- coding: UTF-8 -*-
# @Time    : 2019-06-08 15:51
# @File    : flic.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
import scipy.io as io
from antgo.dataflow.dataset import *
import os
import numpy as np
import cv2
import time
from antgo.framework.helper.fileio.file_client import *


__all__ = ['FLIC']
class FLIC(Dataset):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(FLIC, self).__init__(train_or_test, dir, ext_params)
    
    if not os.path.exists(os.path.join(self.dir, 'examples.mat')):
      if os.environ.get('LOCAL_RANK', 0) == 0:
        if not os.path.exists(os.path.join(self.dir, 'FLIC-full')):
          # 数据集不存在，需要重新下载，并创建标记
          ali = AliBackend()
          ali.download('ali:///dataset/flic/FLIC-full.zip', self.dir)
          os.system(f'cd {self.dir} && unzip FLIC-full.zip')
          os.system('touch DATASET_IS_READY')
        else:
          # 数据集存在，创建标记
          if not os.path.exists('DATASET_IS_READY'):
            os.system('touch DATASET_IS_READY')
      else:
        while True:
          # 等待直到存在指定文件
          if os.path.exists('DATASET_IS_READY'):
            break
          time.sleep(5)

      self.dir = os.path.join(self.dir, 'FLIC-full')

    assert(train_or_test in ['train', 'val'])
    matr = io.loadmat(os.path.join(self.dir, 'examples.mat'))
    dataset = matr['examples'][0]

    self.ids = []
    self.files = []
    self.coords = []
    self.image_size = []
    self.torso = []

    self.valid_coords = [0,1,2,3,4,5,6,9,12,13,16]
    self.coords_name = ['Nose',
                        'Right Shoulder',
                        'Right Elbow',
                        'Right Wrist',
                        'Right Hip',
                        'Left Shoulder',
                        'Left Elbow',
                        'Left Wrist',
                        'Left Hip',
                        'Left Eye',
                        'Right Eye']
    count = 0
    for data in dataset:
      if train_or_test == 'train' and data[7][0][0]:
        # add train set
        self.ids.append(count)
        self.files.append(data[3][0])
        self.coords.append(data[2]) # 2 x 20
        self.image_size.append(data[4])
        self.torso.append(data[6])

        count += 1
      elif train_or_test == 'val' and data[8][0][0]:
        # add test set
        self.ids.append(count)
        self.files.append(data[3][0])
        self.coords.append(data[2]) # 2 x 20
        self.image_size.append(data[4])
        self.torso.append(data[6])

        count += 1

  @property
  def size(self):
    return len(self.ids)

  def at(self, id):
    img = cv2.imread(os.path.join(self.dir, 'images', self.files[id]))
    joints2d = np.transpose(self.coords[id][:,self.valid_coords], (1,0))
    bboxes = np.array([[
      np.min(joints2d[:, 0]), 
      np.min(joints2d[:, 1]), 
      np.max(joints2d[:, 0]), 
      np.max(joints2d[:, 1])
    ]])
    return (
      img, 
      {
        'joints2d': np.expand_dims(joints2d, 0),
        'joints_vis': np.ones((1, len(joints2d))),
        'bboxes': bboxes, 
        'labels': np.ones((1), dtype=np.int32),
        'image_meta': {
          'image_shape': (img.shape[0], img.shape[1])
        }        
      }
    )

  def split(self, split_params={}, split_method=''):
    raise NotImplementedError

# p2012 = FLIC('val', '/root/workspace/dataset/A')
# print(f'p2012 size {p2012.size}')
# for i in range(p2012.size):
#   data = p2012.sample(i)
#   # ss = result['segments']
#   image = data['image']
#   joints2d = data['joints2d']
#   for joint_i, (x,y) in enumerate(joints2d[0]):
#       x, y = int(x), int(y)
#       cv2.circle(image, (x, y), radius=2, color=(0,0,255), thickness=1)
#   cv2.imwrite('./1234.png', image.astype(np.uint8))

#   print(i)
# value = p2012.sample(0)
# print(value.keys())
# value = p2012.sample(1)
# print(value)


# 模型配置
model = dict(
    type='ACModule',
    test_cfg=dict(
        feature_config=dict(
            from_name='backbone',
            from_index=-1
        ),
    ),
    train_cfg=None,
    init_cfg=dict(
        sampling_num=5000,                          # 采样总数
        sampling_count=5,
        sampling_fn=dict(
            type='SamplingByComposer',
            sampling_cfgs=[
                dict(
                    type='SamplingByUniform',
                    sampling_num=10000              # 随机采样总数 (第一步，随机采样10000)
                ),
                dict(
                    type='SamplingByKCenter',
                    sampling_num=5000,              # 聚类采样总数 (在第二步，从上一步子集中采样5000)
                    key='feature'
                )
            ]
        ),             
    ),
)



from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

__graph_info = {}

def add_op_info(op_name, op_index, op_args, op_kwargs):
    global __graph_info
    if 'op' not in __graph_info:
        __graph_info['op'] = []
    __graph_info['op'].append({
        'op_name': op_name,     # 算子名称
        'op_index': op_index,   # 上下游数据流转
        'op_args': op_args,     # 算子参数
        'op_kwargs': op_kwargs  # 算子参数
    })


def add_input_config(input_config):
    global __graph_info
    __graph_info['input'] = input_config


def add_output_config(output_config):
    global __graph_info
    __graph_info['output'] = output_config


def get_graph_info():
    global __graph_info
    return __graph_info


from antgo.framework.helper.runner.hooks.hook import HOOKS, Hook

# @HOOKS.register_module()
# class YourHook(Hook):
#     def __init__(self, *args, **kwargs):
#         super().__init__()
    
#     def before_run(self, runner):
#         # 在全局迭代前，运行一次
#         pass
#     def after_run(self, runner):
#         # 在全局推出前，运行一次
#         pass
    
#     def before_train_epoch(self, runner):
#         # 在每个epoch前，运行一次
#         pass
#     def after_train_epoch(self, runner):
#         # 在每个epoch后，运行一次
#         pass
    
#     def before_train_iter(self, runner):
#         # 在每次迭代前，运行一次
#         pass
#     def after_train_iter(self, runner):
#         # 在每次迭代后，运行一次
#         pass


"""
Implements ADDA:
Adversarial Discriminative Domain Adaptation, Tzeng et al. (2017)
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.models.builder import MODELS
from antgo.framework.helper.runner import BaseModule
from antgo.framework.helper.utils import Registry, build_from_cfg
from antgo.framework.helper.runner.optimizer.builder import build_optimizer
from antgo.framework.helper.models.builder import build_model
from antgo.framework.helper.runner.hooks.hook import *


class Discriminator(torch.nn.Module):
    def __init__(self, input_dim=1024, latent_dim=512):
        super(Discriminator, self).__init__()
        self.input_dim = input_dim
        self.latent_dim = latent_dim

        self.model = torch.nn.Sequential(
            torch.nn.Linear(self.input_dim, self.latent_dim),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Linear(self.latent_dim, self.latent_dim),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Linear(self.latent_dim, 1),
            torch.nn.Sigmoid(),
        )

    def forward(self, z):
        validity = self.model(z).view(-1)
        return validity


@MODELS.register_module()
class Adda(BaseModule):
    def __init__(self, model: dict, train_cfg=None, test_cfg=None, init_cfg=None):
        super(Adda, self).__init__(init_cfg=init_cfg)

        # model build (具体任务模型)
        self.model = build_model(model)

        # discriminator (判别器模型)
        assert(train_cfg.get('latent_dim'))
        self.discriminator = Discriminator(train_cfg.input_dim, train_cfg.latent_dim)
        self.dis_loss_func = torch.nn.BCELoss()

        self.src_domain_batch_size = train_cfg.get('src_domain_batch_size', 5)      # 目标域数据量 在一个batch里
        self.gen_loss_weight = train_cfg.get('gen_loss_weight', 1.0)                # 生成器损失权重
        self.dis_loss_weight = train_cfg.get('dis_loss_weight', 1.0)                # 判别器损失权重

        self.encoder_config = train_cfg.encoder_config  # from_name='', from_index=-1       

    def train_step(self, data, optimizer, **kwargs):
        # -----------------
        #  STEP 1: Train Generator
        # -----------------        
        # 下游任务处于train状态 + 判别器处于eval状态
        self.model.train()
        self.discriminator.eval()
        optimizer['model'].zero_grad()

        num_samples = 1
        if type(data) == list or type(data) == tuple:
            num_samples = len(data[0]['image'])
        else:
            num_samples = len(data['image'])

        encoder_feature_proxy = getattr(getattr(self.model, self.encoder_config.from_name), 'forward', None)
        from_index = self.encoder_config.from_index
        out = {}
        def feature_extract_func(x):
            x = encoder_feature_proxy(x)
            feature = x
            if isinstance(x, tuple):
                feature = x[from_index]
            batch_size = feature.shape[0]
            out.update({
                'feature': feature.view(batch_size, -1)
            })
            return x
        setattr(getattr(self.model, self.encoder_config.from_name), 'forward', feature_extract_func)

        # 下游任务模型损失
        step_1_losses = self.model(**data, **kwargs)

        src_domain_batch_size = self.src_domain_batch_size
        tgt_domain_batch_size = num_samples - self.src_domain_batch_size
        src_domain_encoder_feature = out['feature'][:self.src_domain_batch_size]       # 例如，仿真数据 (0)
        tgt_domain_encoder_feature = out['feature'][self.src_domain_batch_size:]       # 例如，真实数据 (1)

        # 
        gen_loss = self.dis_loss_func(
            self.discriminator(src_domain_encoder_feature), 
            torch.Tensor(src_domain_encoder_feature.shape[0]).fill_(1.0).to(src_domain_encoder_feature.device))
        step_1_losses.update(
            {
                'adv_loss': self.gen_loss_weight*gen_loss
            }
        )

        step_1_total_loss, step_1_log_vars = self._parse_losses(step_1_losses)
        step_1_total_loss.backward()
        optimizer['model'].step()    # 仅影响生成器

        # ---------------------
        #  STEP 2: Train Discriminator
        # ---------------------
        self.model.eval()
        self.discriminator.train()
        optimizer['discriminator'].zero_grad()

        d_src_loss = self.dis_loss_func(
            self.discriminator(src_domain_encoder_feature.detach()), 
            torch.Tensor(src_domain_encoder_feature.shape[0]).fill_(0.0).to(src_domain_encoder_feature.device))
        d_tgt_loss = self.dis_loss_func(
            self.discriminator(tgt_domain_encoder_feature.detach()), 
            torch.Tensor(tgt_domain_encoder_feature.shape[0]).fill_(1.0).to(tgt_domain_encoder_feature.device))

        # TODO，训练判别器，保证真实样本损失和仿真样本损失 平衡
        src_weight = tgt_domain_batch_size / src_domain_batch_size
        tgt_weight = 1.0
        step_2_losses = {
            'd_src_loss': self.dis_loss_weight * d_src_loss*src_weight,
            'd_tgt_loss': self.dis_loss_weight * d_tgt_loss*tgt_weight
        }
        step_2_total_loss, step_2_log_vars = self._parse_losses(step_2_losses)

        step_2_total_loss.backward()
        optimizer['discriminator'].step()   # 仅影响判别器
        
        # 恢复原有设置
        setattr(getattr(self.model, self.encoder_config.from_name), 'forward', encoder_feature_proxy)

        # 结束
        log_vars = step_1_log_vars
        log_vars.update(step_2_log_vars)
        outputs = dict(
            loss=0.0, log_vars=log_vars, num_samples=num_samples)
        return outputs


# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
from antgo.framework.helper.utils import Registry, build_from_cfg

TRANSFORMER = Registry('Transformer')
LINEAR_LAYERS = Registry('linear layers')


def build_transformer(cfg, default_args=None):
    """Builder for Transformer."""
    return build_from_cfg(cfg, TRANSFORMER, default_args)


LINEAR_LAYERS.register_module('Linear', module=nn.Linear)


def build_linear_layer(cfg, *args, **kwargs):
    """Build linear layer.
    Args:
        cfg (None or dict): The linear layer config, which should contain:
            - type (str): Layer type.
            - layer args: Args needed to instantiate an linear layer.
        args (argument list): Arguments passed to the `__init__`
            method of the corresponding linear layer.
        kwargs (keyword arguments): Keyword arguments passed to the `__init__`
            method of the corresponding linear layer.
    Returns:
        nn.Module: Created linear layer.
    """
    if cfg is None:
        cfg_ = dict(type='Linear')
    else:
        if not isinstance(cfg, dict):
            raise TypeError('cfg must be a dict')
        if 'type' not in cfg:
            raise KeyError('the cfg dict must contain the key "type"')
        cfg_ = cfg.copy()

    layer_type = cfg_.pop('type')
    if layer_type not in LINEAR_LAYERS:
        raise KeyError(f'Unrecognized linear type {layer_type}')
    else:
        linear_layer = LINEAR_LAYERS.get(layer_type)

    layer = linear_layer(*args, **kwargs, **cfg_)

    return layer



from .classification import *
from .semi import *
from .detectors import *
from .nas import *
from .activelearning import *
from .adda import *
from .pose3d import *
from .segmentation import *
from .distillation import *


# Copyright (c) OpenMMLab. All rights reserved.
import copy

from ..utils import Registry

RUNNERS = Registry('runner')
RUNNER_BUILDERS = Registry('runner builder')

MEASURES = Registry("measures")

def build_runner_constructor(cfg):
    return RUNNER_BUILDERS.build(cfg)


def build_runner(cfg, default_args=None):
    runner_cfg = copy.deepcopy(cfg)
    constructor_type = runner_cfg.pop('constructor',
                                      'DefaultRunnerConstructor')
    runner_constructor = build_runner_constructor(
        dict(
            type=constructor_type,
            runner_cfg=runner_cfg,
            default_args=default_args))
    runner = runner_constructor()
    return runner

def build_measure(cfg):
    measure_obj = MEASURES.build(cfg)
    return measure_obj


# -*- coding: UTF-8 -*-
# @Time    : 18-3-14
# @File    : ali_fashionai_landmark.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np

__all__ = ['FashionAILandmark']
class FashionAILandmark(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(FashionAILandmark, self).__init__(train_or_test, dir)
    
    assert(train_or_test in ['train', 'test'])
    if train_or_test == 'train':
      if not (os.path.exists(os.path.join(self.dir,'train','Images')) and
                os.path.exists(os.path.join(self.dir, 'train','Annotations'))):
        logger.error('FashionAILandmark train dataset must download from \n https://tianchi.aliyun.com/competition/information.htm?raceId=231648')
        sys.exit(0)
    else:
      if not (os.path.exists(os.path.join(self.dir, 'test','Images')) and
                os.path.exists(os.path.join(self.dir, 'test','test.csv'))):
        logger.error('FashionAILandmark test dataset must download from \n https://tianchi.aliyun.com/competition/information.htm?raceId=231648')
        sys.exit(0)
    
    self.key_points = ['neckline_left',     #左领部
                      'neckline_right',     #右领部
                      'center_front',       #中线
                      'shoulder_left',      #左肩部
                      'shoulder_right',     #右肩部
                      'armpit_left',        #左腋窝
                      'armpit_right',       #右腋窝
                      'waistline_left',     #左腰部
                      'waistline_right',    #右腰部
                      'cuff_left_in',       #左袖口内
                      'cuff_left_out',      #左袖口外
                      'cuff_right_in',      #右袖口内
                      'cuff_right_out',     #右袖口外
                      'top_hem_left',       #左衣摆
                      'top_hem_right',      #右衣摆
                      'waistband_left',     #左腰部
                      'waistband_right',    #右腰部
                      'hemline_left',       #左裙摆
                      'hemline_right',      #右裙摆
                      'crotch',             #裆部
                      'bottom_left_in',     #左裤脚内
                      'bottom_left_out',    #左裤脚外
                      'bottom_right_in',    #右裤脚内
                      'bottom_right_out',   #右裤脚外
                      ]

    self.category_landmark = {'blouse': {
                                0:0,
                                1:1,
                                2:2,
                                3:3,
                                4:4,
                                5:5,
                                6:6,
                                9:7,
                                10:8,
                                11:9,
                                12:10,
                                13:11,
                                14:12},
                              'dress': {
                                0:0,
                                1:1,
                                2:2,
                                3:3,
                                4:4,
                                5:5,
                                6:6,
                                7:7,
                                8:8,
                                9:9,
                                10:10,
                                11:11,
                                12:12,
                                17:13,
                                18:14
                              },
                              'outwear': {
                                0:0,
                                1:1,
                                3:2,
                                4:3,
                                5:4,
                                6:5,
                                7:6,
                                8:7,
                                9:8,
                                10:9,
                                11:10,
                                12:11,
                                13:12,
                                14:13
                              },
                              'skirt': {
                                15:0,
                                16:1,
                                17:2,
                                18:3,
                              },
                              'trousers': {
                                15:0,
                                16:1,
                                19:2,
                                20:3,
                                21:4,
                                22:5,
                                23:6
                              }}

    self.category_map = {
      'blouse': 0,
      'dress': 1,
      'outwear': 2,
      'skirt': 3,
      'trousers': 4,
    }
    
    self.annotation = []
    self.images = []
    if train_or_test == 'train':
      with open(os.path.join(self.dir, 'train', 'Annotations', 'train.csv')) as fp:
        # skip first row
        content = fp.readline()
        content = fp.readline()
        while content:
          key_terms = content.split(',')
          image_id = key_terms[0]
          category = key_terms[1]
          # record image id
          self.images.append((image_id, category))

          # record annotation
          sample_annotation = {}
          sample_annotation['category'] = category
          sample_annotation['category_id'] = self.category_map[category]
          sample_annotation['landmark'] = []
          sample_annotation['id'] = len(self.images)
          key_point_annotation = key_terms[2:]
          for kp_index, kp in enumerate(key_point_annotation):
            x, y, visible = kp.split('_')
            x = int(x)
            y = int(y)
            visible = int(visible)

            if kp_index in self.category_landmark[category]:
              sample_annotation['landmark'].append((self.category_landmark[category][kp_index], kp_index, x, y, visible))

          self.annotation.append(sample_annotation)

          # read next line
          content = fp.readline()

      with open(os.path.join(self.dir, 'train', 'Annotations', 'annotations.csv')) as fp:
        # skip first row
        content = fp.readline()
        content = fp.readline()
        while content:
          key_terms = content.split(',')
          image_id = key_terms[0]
          category = key_terms[1]
          # record image id
          self.images.append((image_id, category))
    
          # record annotation
          sample_annotation = {}
          sample_annotation['category'] = category
          sample_annotation['category_id'] = self.category_map[category]
          sample_annotation['landmark'] = []
          sample_annotation['id'] = len(self.images)
          key_point_annotation = key_terms[2:]
          for kp_index, kp in enumerate(key_point_annotation):
            x, y, visible = kp.split('_')
            x = int(x)
            y = int(y)
            visible = int(visible)
      
            if kp_index in self.category_landmark[category]:
              sample_annotation['landmark'].append(
                (self.category_landmark[category][kp_index], kp_index, x, y, visible))
    
          self.annotation.append(sample_annotation)
    
          # read next line
          content = fp.readline()
    else:
      with open(os.path.join(self.dir, 'test', 'test.csv')) as fp:
        # skip first row
        content = fp.readline()
        content = fp.readline()
        content = content.replace('\n','')
        while content:
          key_terms = content.split(',')
          image_id = key_terms[0]
          category = key_terms[1]
          # record image id
          self.images.append((image_id, category))

          # read next line
          content = fp.readline()
          content = content.replace('\n','')
    
    # data index list
    self.ids = list(range(len(self.images)))
    # fixed seed
    self.seed = 0
    
  @property
  def size(self):
    return len(self.ids)
  
  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1
    
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)
    
      for k in idxs:
        image_file, category = self.images[k]
        image_path = os.path.join(self.dir, self.train_or_test, image_file)
        image = imread(image_path)
        
        if self.train_or_test == 'train':
          data_annotation = copy.deepcopy(self.annotation[k])
          data_annotation['info'] = [image.shape[0], image.shape[1], image.shape[2]]
          yield [(image, category, image_file), data_annotation]
        else:
          yield [(image, category, image_file), None]

  def at(self, id):
    image_file, category_id = self.images[id]
    image_path = os.path.join(self.dir, self.train_or_test, image_file)
    image = imread(image_path)
    return (image, category_id, image_file)

  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold', 'holdout'])

    # set fixed random seed
    np.random.seed(np.int64(self.seed))
    category_ids = None
    if split_method == 'kfold':
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [0 for _ in range(len(self.ids))]

    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = FashionAILandmark(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids

    val_dataset = FashionAILandmark(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids

    return train_dataset, val_dataset  # split data by their label



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import os.path as osp
import ast
import astor
from x2paddle.project_convertor.pytorch.mapper import *
import copy
from .utils import get_dep_file_path


class PtDepInfo:
    """
    PyTorch依赖包信息。
    FROM代表from信息的字符串，例如：torch；
    IMPORT代表import信息系的字符串，例如：nn.functional；
    AS代表as信息的字符串，例如：F；
    DEPENDENCY代表由FROM、IMPORT、AS三者组成的字符串，例如：from torch import nn.functional as F。
    """
    FROM = None
    IMPORT = None
    AS = None
    DEPENDENCY = None


class DependencyAnalyzer(ast.NodeVisitor):
    """ 获取python文件的依赖信息。
        依赖信息由4部分组成：（1）import相关的依赖；（2）赋值；（3）函数；（4）类。

    Args:
        py_file_path (str): python文件的绝对值路径。
        file_dependencies (dict): 当前已经统计的依赖信息，key为python文件的绝对值路径，
                                value为key文件所对应的依赖信息组成的list。
    """

    def __init__(self, py_file_path, file_dependencies):
        self.py_file_path = py_file_path
        self.file_dependencies = file_dependencies
        self.root = ast.parse(open(py_file_path, "rb").read())
        self.scopes_and_dependencies = list()  # 作用域和依赖组成的stack
        self.file_dependencies[self.py_file_path] = list()

    def _get_scope_node(self):
        """ 获取当前节点的作用域。
        """
        scope_node = None
        for i in range(len(self.scopes_and_dependencies)):
            i = -(i + 1)
            sd = self.scopes_and_dependencies[i]
            if not isinstance(sd, PtDepInfo) and not isinstance(sd, ast.Assign):
                scope_node = sd
                break
        return scope_node

    def run(self):
        self.scopes_and_dependencies.append(self.root)
        self.visit(self.root)

    def visit(self, node):
        out = super(DependencyAnalyzer, self).visit(node)

    def visit_ImportFrom(self, node):
        """ 遍历子节点。
        """
        son_nodes = node.names
        for son_node in son_nodes:
            self.visit_alias(son_node, node.module, node.level)

    def visit_Import(self, node):
        """ 遍历子节点。
        """
        son_nodes = getattr(node, "names")
        for son_node in son_nodes:
            self.visit_alias(son_node)

    def visit_alias(self, node, from_name=None, from_level=None):
        """ 构建PtDepInfo并将当前的PtDepInfo放入scopes_and_dependencies。
            如果import字符串为“*”，获取依赖包所在文件的依赖信息加入当前的dependencies；
            反之，直接在dependencies中加入PtDepInfo。
        """
        dep_info = PtDepInfo()
        dep_info.FROM = from_name
        dep_info.IMPORT = getattr(node, "name")
        dep_info.AS = getattr(node, "asname", None)
        if dep_info.IMPORT == "*":
            import_file_path = get_dep_file_path(self.py_file_path, from_level,
                                                 dep_info.FROM)
            if import_file_path not in self.file_dependencies:
                analyzer = DependencyAnalyzer(import_file_path,
                                              self.file_dependencies)
                analyzer.run()
            self.file_dependencies[self.py_file_path].extend(
                self.file_dependencies[import_file_path])
        else:
            dependency_str_list = list()
            if dep_info.FROM is None and from_level is not None:
                dependency_str_list.append("." * from_level)
            elif dep_info.FROM is not None:
                dependency_str_list.append(dep_info.FROM)
            dependency_str_list.append(dep_info.IMPORT)
            dep_info.DEPENDENCY = ".".join(dependency_str_list)
            self.file_dependencies[self.py_file_path].append(dep_info)
        self.scopes_and_dependencies.append(dep_info)

    def visit_FunctionDef(self, node):
        """ 当作用域为ast的根节点时，把函数名放入dependencies。
        """
        if isinstance(self._get_scope_node(), ast.Module):
            self.scopes_and_dependencies.append(node)
            self.file_dependencies[self.py_file_path].append(node.name)

    def visit_ClassDef(self, node):
        """ 当作用域为ast的根节点时，把类名放入dependencies。
        """
        if isinstance(self._get_scope_node(), ast.Module):
            self.scopes_and_dependencies.append(node)
            self.file_dependencies[self.py_file_path].append(node.name)

    def visit_Assign(self, node):
        """ 当作用域为ast的根节点时，把赋值名放入dependencies。
        """
        if isinstance(self._get_scope_node(), ast.Module):
            self.scopes_and_dependencies.append(node)
            for target in node.targets:
                if isinstance(target, ast.Tuple):
                    for ele in target.elts:
                        self.file_dependencies[self.py_file_path].append(ele.id)
                elif isinstance(target, ast.Name):
                    self.file_dependencies[self.py_file_path].append(target.id)


def analyze(py_file_path, file_dependencies):
    analyzer = DependencyAnalyzer(py_file_path, file_dependencies)
    analyzer.run()



from abc import ABCMeta, abstractmethod

import torch
from antgo.framework.helper.runner import BaseModule
from antgo.framework.helper.models.builder import *

class ProxyModule(BaseModule):
    def __init__(self, feat_module, head_cfg, train_cfg=None, test_cfg=None, init_cfg=None) -> None:
        super().__init__(init_cfg)
        if train_cfg is not None:
            head_cfg.update({'train_cfg':train_cfg})
        if test_cfg is not None:
            head_cfg.update({'test_cfg':test_cfg})

        head_init_cfg = dict()
        if 'init_cfg' in head_cfg and head_cfg['init_cfg'] is not None:
            head_init_cfg = head_cfg['init_cfg']
        head_init_cfg['compute_feature'] = False
        head_cfg.update({
            'init_cfg': head_init_cfg
        })

        self.head_module = build_head(head_cfg)
        self.feat_module = feat_module

    def forward(self, *data, **kwargs):
        """Calls either :func:`forward_train` or :func:`forward_test` depending
        on whether ``return_loss`` is ``True``.

        Note this setting will change the expected inputs. When
        ``return_loss=True``, img and img_meta are single-nested (i.e. Tensor
        and List[dict]), and when ``resturn_loss=False``, img and img_meta
        should be double nested (i.e.  List[Tensor], List[List[dict]]), with
        the outer list indicating test time augmentations.
        """
        return_loss = True
        if 'return_loss' in kwargs:
            return_loss = kwargs['return_loss']

        if return_loss:
            return self.forward_train(*data, **kwargs)
        else:
            return self.forward_test(*data, **kwargs)

    def forward_test(self, *data, **kwargs):
        feat = self.feat_module(kwargs['image'])
        kwargs.update({
            'feature': feat
        })
 
        results_list = self.head_module.simple_test(
            None, **kwargs)

        results = results_list
        if self.init_cfg.get('task', None) != None:
            if self.init_cfg.get('task') == 'det':
                results = {
                    'box': torch.stack([a for a, _ in results_list], dim=0),
                    'label': torch.stack([b for _, b in results_list], dim=0)
                }
        return results

    def forward_train(self, *data, **kwargs):
        feat = self.feat_module(kwargs['image'])
        kwargs.update({
            'feature': feat
        })

        return self.head_module.forward_train(None, **kwargs)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd.gradcheck import zero_gradients
from ...builder import DISTILL_LOSSES
'''
Modified by https://github.com/bhheo/BSS_distillation
'''


@DISTILL_LOSSES.register_module()
def reduce_sum(x, keepdim=True):
	for d in reversed(range(1, x.dim())):
		x = x.sum(d, keepdim=keepdim)
	return x


def l2_norm(x, keepdim=True):
	norm = reduce_sum(x*x, keepdim=keepdim)
	return norm.sqrt()


class BSS(nn.Module):
	'''
	Knowledge Distillation with Adversarial Samples Supporting Decision Boundary
	https://arxiv.org/pdf/1805.05532.pdf
	'''
	def __init__(self, T):
		super(BSS, self).__init__()
		self.T = T

	def forward(self, attacked_out_s, attacked_out_t):
		loss = F.kl_div(F.log_softmax(attacked_out_s/self.T, dim=1),
						F.softmax(attacked_out_t/self.T, dim=1),
						reduction='batchmean') #* self.T * self.T

		return loss


class BSSAttacker():
	def __init__(self, step_alpha, num_steps, eps=1e-4):
		self.step_alpha = step_alpha
		self.num_steps = num_steps
		self.eps = eps

	def attack(self, model, img, target, attack_class):
		img = img.detach().requires_grad_(True)

		step = 0
		while step < self.num_steps:
			zero_gradients(img)
			_, _, _, _, _, output = model(img)

			score = F.softmax(output, dim=1)
			score_target = score.gather(1, target.unsqueeze(1))
			score_attack_class = score.gather(1, attack_class.unsqueeze(1))

			loss = (score_attack_class - score_target).sum()
			loss.backward()

			step_alpha = self.step_alpha * (target == output.max(1)[1]).float()
			step_alpha = step_alpha.unsqueeze(1).unsqueeze(1).unsqueeze(1)
			if step_alpha.sum() == 0:
				break

			pert = (score_target - score_attack_class).unsqueeze(1).unsqueeze(1)
			norm_pert = step_alpha * (pert + self.eps) * img.grad / l2_norm(img.grad)

			step_adv = img + norm_pert
			step_adv = torch.clamp(step_adv, -2.5, 2.5)
			img.data = step_adv.data

			step += 1

		return img


import torch.nn as nn
import torch.nn.functional as F

from antgo.framework.helper.models.builder import LOSSES
from .utils import weight_reduce_loss


def varifocal_loss(pred,
                   target,
                   weight=None,
                   alpha=0.75,
                   gamma=2.0,
                   iou_weighted=True,
                   reduction='mean',
                   avg_factor=None):
    """`Varifocal Loss <https://arxiv.org/abs/2008.13367>`_

    Args:
        pred (torch.Tensor): The prediction with shape (N, C), C is the
            number of classes
        target (torch.Tensor): The learning target of the iou-aware
            classification score with shape (N, C), C is the number of classes.
        weight (torch.Tensor, optional): The weight of loss for each
            prediction. Defaults to None.
        alpha (float, optional): A balance factor for the negative part of
            Varifocal Loss, which is different from the alpha of Focal Loss.
            Defaults to 0.75.
        gamma (float, optional): The gamma for calculating the modulating
            factor. Defaults to 2.0.
        iou_weighted (bool, optional): Whether to weight the loss of the
            positive example with the iou target. Defaults to True.
        reduction (str, optional): The method used to reduce the loss into
            a scalar. Defaults to 'mean'. Options are "none", "mean" and
            "sum".
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
    """
    # pred and target should be of the same size
    assert pred.size() == target.size()
    pred_sigmoid = pred.sigmoid()
    target = target.type_as(pred)
    if iou_weighted:
        focal_weight = target * (target > 0.0).float() + \
            alpha * (pred_sigmoid - target).abs().pow(gamma) * \
            (target <= 0.0).float()
    else:
        focal_weight = (target > 0.0).float() + \
            alpha * (pred_sigmoid - target).abs().pow(gamma) * \
            (target <= 0.0).float()
    loss = F.binary_cross_entropy_with_logits(
        pred, target, reduction='none') * focal_weight
    loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
    return loss


@LOSSES.register_module()
class VarifocalLoss(nn.Module):

    def __init__(self,
                 use_sigmoid=True,
                 alpha=0.75,
                 gamma=2.0,
                 iou_weighted=True,
                 reduction='mean',
                 loss_weight=1.0):
        """`Varifocal Loss <https://arxiv.org/abs/2008.13367>`_

        Args:
            use_sigmoid (bool, optional): Whether the prediction is
                used for sigmoid or softmax. Defaults to True.
            alpha (float, optional): A balance factor for the negative part of
                Varifocal Loss, which is different from the alpha of Focal
                Loss. Defaults to 0.75.
            gamma (float, optional): The gamma for calculating the modulating
                factor. Defaults to 2.0.
            iou_weighted (bool, optional): Whether to weight the loss of the
                positive examples with the iou target. Defaults to True.
            reduction (str, optional): The method used to reduce the loss into
                a scalar. Defaults to 'mean'. Options are "none", "mean" and
                "sum".
            loss_weight (float, optional): Weight of loss. Defaults to 1.0.
        """
        super(VarifocalLoss, self).__init__()
        assert use_sigmoid is True, \
            'Only sigmoid varifocal loss supported now.'
        assert alpha >= 0.0
        self.use_sigmoid = use_sigmoid
        self.alpha = alpha
        self.gamma = gamma
        self.iou_weighted = iou_weighted
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None):
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction.
            target (torch.Tensor): The learning target of the prediction.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Options are "none", "mean" and "sum".

        Returns:
            torch.Tensor: The calculated loss
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if self.use_sigmoid:
            loss_cls = self.loss_weight * varifocal_loss(
                pred,
                target,
                weight,
                alpha=self.alpha,
                gamma=self.gamma,
                iou_weighted=self.iou_weighted,
                reduction=reduction,
                avg_factor=avg_factor)
        else:
            raise NotImplementedError
        return loss_cls



# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File: cifar.py
# @Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import os, sys
import pickle
import numpy as np
import random
import six
from six.moves import range
from ...utils import logger
from ...utils.fs import download
from ...utils.fs import maybe_here
from .dataset import Dataset
import copy
import time

__all__ = ['Cifar10', 'Cifar100']
CIFAR_10_URL = 'http://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'
CIFAR_100_URL = 'http://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz'


def read_cifar(filenames, cifar_classnum):
  assert cifar_classnum == 10 or cifar_classnum == 100
  ret = []
  for fname in filenames:
    fo = open(fname, 'rb')
    if six.PY3:
      dic = pickle.load(fo, encoding='bytes')
    else:
      dic = pickle.load(fo)
    data = dic[b'data']
    if cifar_classnum == 10:
      label = dic[b'labels']
      IMG_NUM = 10000 # cifar10 data are split into blocks of 10000
    elif cifar_classnum == 100:
      label = dic[b'fine_labels']
      IMG_NUM = 50000 if 'train' in fname else 10000
    fo.close()
    for k in range(IMG_NUM):
      img = data[k].reshape(3, 32, 32)
      img = np.transpose(img, [1, 2, 0])
      ret.append([img, label[k]])
  return ret


def get_filenames(dir, cifar_classnum):
  assert cifar_classnum == 10 or cifar_classnum == 100
  filenames = []
  if cifar_classnum == 10:
    filenames = [os.path.join(dir, 'cifar-10-batches-py', 'data_batch_%d' % i) for i in range(1, 6)]
    filenames.append(os.path.join(dir, 'cifar-10-batches-py', 'test_batch'))
  elif cifar_classnum == 100:
    filenames = [os.path.join(dir, 'cifar-100-python', 'train'),
                 os.path.join(dir, 'cifar-100-python', 'test')]
  return filenames


class CifarBase(Dataset):
  """
  Return [image, label],
      image is 32x32x3 in the range [0,255]
  """
  def __init__(self, train_or_test, dir=None, params=None, cifar_classnum=10):
    assert cifar_classnum == 10 or cifar_classnum == 100
    super(CifarBase,self).__init__(train_or_test,dir)
    self.train_or_test = train_or_test
    self.cifar_classnum = cifar_classnum
    self.dir = dir

    if not os.path.exists(self.dir):
      os.makedirs(self.dir)

    # fixed seed
    self.seed = time.time()

    """Download and extract the tarball from Alex's website."""
    cifar_foldername = ''
    if cifar_classnum == 10:
      cifar_foldername = 'cifar-10-batches-py'
    else:
      cifar_foldername = 'cifar-100-python'

    data_url = CIFAR_10_URL if cifar_classnum == 10 else CIFAR_100_URL
    if os.environ.get('LOCAL_RANK', 0) == 0:
      if not os.path.exists(os.path.join(self.dir, cifar_foldername)):
        # 数据集不存在，需要重新下载，并创建标记
        self.download(self.dir, default_url=data_url, auto_untar=True, is_gz=True)
        os.system('touch DATASET_IS_READY')     
      else:
        # 数据集存在，创建标记
        if not os.path.exists('DATASET_IS_READY'):
          os.system('touch DATASET_IS_READY')
    else:
      while True:
        # 等待直到存在指定文件
        if os.path.exists('DATASET_IS_READY'):
          break
        time.sleep(5)

    fnames = get_filenames(self.dir, self.cifar_classnum)
    if self.train_or_test == 'train':
      self.fs = fnames[:-1]
    else:
      self.fs = [fnames[-1]]
    for f in self.fs:
      if not os.path.isfile(f):
        raise ValueError('Failed to Find File: ' + f)
    self.train_or_test = self.train_or_test
    self.data = read_cifar(self.fs, self.cifar_classnum)

    # unwarp
    self.image, self.label = zip(*self.data)

    # ids
    self.ids = [i for i in range(len(self.image))]

  @property
  def size(self):
    return len(self.ids)

  def get_cat_ids(self, idx):
    return self.label[idx]
  
  def get_ann_info(self, idx):
    return {'label': self.label[idx]}
  
  def at(self, id):
    return self.image[id], self.label[id]


class Cifar10(CifarBase):
  def __init__(self, train_or_test, dir=None, params=None):
    super(Cifar10, self).__init__(train_or_test, dir, params,10)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold'])

    category_ids = None
    if split_method == 'kfold':
      np.random.seed(np.int64(self.seed))
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [self.label[i] for i in range(len(self.ids))]

    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = Cifar10(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids

    val_dataset = Cifar10(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids

    return train_dataset, val_dataset  # split data by their label


class Cifar100(CifarBase):
  def __init__(self, train_or_test, dir=None, params=None):
    super(Cifar100, self).__init__(train_or_test, dir, params, 100)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold'])

    category_ids = None
    if split_method == 'kfold':
      np.random.seed(np.int64(self.seed))
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [self.label[i] for i in range(len(self.ids))]

    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = Cifar100(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids

    val_dataset = Cifar100(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids

    return train_dataset, val_dataset  # split data by their label



from __future__ import unicode_literals
from __future__ import division

import numpy as np

def confusion_matrix(rater_a, rater_b, min_rating=None, max_rating=None):
    """
    Returns the confusion matrix between rater's ratings
    """
    assert(len(rater_a) == len(rater_b))
    if min_rating is None:
        min_rating = min(rater_a + rater_b)
    if max_rating is None:
        max_rating = max(rater_a + rater_b)
    num_ratings = int(max_rating - min_rating + 1)
    conf_mat = [[0 for i in range(num_ratings)]
                for j in range(num_ratings)]
    for a, b in zip(rater_a, rater_b):
        conf_mat[a - min_rating][b - min_rating] += 1
    return conf_mat


def histogram(ratings, min_rating=None, max_rating=None):
    """
    Returns the counts of each type of rating that a rater made
    """
    if min_rating is None:
        min_rating = min(ratings)
    if max_rating is None:
        max_rating = max(ratings)
    num_ratings = int(max_rating - min_rating + 1)
    hist_ratings = [0 for x in range(num_ratings)]
    for r in ratings:
        hist_ratings[r - min_rating] += 1
    return hist_ratings


def quadratic_weighted_kappa(rater_a, rater_b, min_rating=None, max_rating=None):
    """
    Calculates the quadratic weighted kappa
    quadratic_weighted_kappa calculates the quadratic weighted kappa
    value, which is a measure of inter-rater agreement between two raters
    that provide discrete numeric ratings.  Potential values range from -1
    (representing complete disagreement) to 1 (representing complete
    agreement).  A kappa value of 0 is expected if all agreement is due to
    chance.
    quadratic_weighted_kappa(rater_a, rater_b), where rater_a and rater_b
    each correspond to a list of integer ratings.  These lists must have the
    same length.
    The ratings should be integers, and it is assumed that they contain
    the complete range of possible ratings.
    quadratic_weighted_kappa(X, min_rating, max_rating), where min_rating
    is the minimum possible rating, and max_rating is the maximum possible
    rating
    """
    rater_a = np.array(rater_a, dtype=int)
    rater_b = np.array(rater_b, dtype=int)
    assert(len(rater_a) == len(rater_b))
    if min_rating is None:
        min_rating = min(min(rater_a), min(rater_b))
    if max_rating is None:
        max_rating = max(max(rater_a), max(rater_b))
    conf_mat = confusion_matrix(rater_a, rater_b,
                                min_rating, max_rating)
    num_ratings = len(conf_mat)
    num_scored_items = float(len(rater_a))

    hist_rater_a = histogram(rater_a, min_rating, max_rating)
    hist_rater_b = histogram(rater_b, min_rating, max_rating)

    numerator = 0.0
    denominator = 0.0

    for i in range(num_ratings):
        for j in range(num_ratings):
            expected_count = (hist_rater_a[i] * hist_rater_b[j]
                              / num_scored_items)
            d = pow(i - j, 2.0) / pow(num_ratings - 1, 2.0)
            numerator += d * conf_mat[i][j] / num_scored_items
            denominator += d * expected_count / num_scored_items

    return 1.0 - numerator / denominator


def linear_weighted_kappa(rater_a, rater_b, min_rating=None, max_rating=None):
    """
    Calculates the linear weighted kappa
    linear_weighted_kappa calculates the linear weighted kappa
    value, which is a measure of inter-rater agreement between two raters
    that provide discrete numeric ratings.  Potential values range from -1
    (representing complete disagreement) to 1 (representing complete
    agreement).  A kappa value of 0 is expected if all agreement is due to
    chance.
    linear_weighted_kappa(rater_a, rater_b), where rater_a and rater_b
    each correspond to a list of integer ratings.  These lists must have the
    same length.
    The ratings should be integers, and it is assumed that they contain
    the complete range of possible ratings.
    linear_weighted_kappa(X, min_rating, max_rating), where min_rating
    is the minimum possible rating, and max_rating is the maximum possible
    rating
    """
    assert(len(rater_a) == len(rater_b))
    if min_rating is None:
        min_rating = min(rater_a + rater_b)
    if max_rating is None:
        max_rating = max(rater_a + rater_b)
    conf_mat = confusion_matrix(rater_a, rater_b,
                                min_rating, max_rating)
    num_ratings = len(conf_mat)
    num_scored_items = float(len(rater_a))

    hist_rater_a = histogram(rater_a, min_rating, max_rating)
    hist_rater_b = histogram(rater_b, min_rating, max_rating)

    numerator = 0.0
    denominator = 0.0

    for i in range(num_ratings):
        for j in range(num_ratings):
            expected_count = (hist_rater_a[i] * hist_rater_b[j]
                              / num_scored_items)
            d = abs(i - j) / float(num_ratings - 1)
            numerator += d * conf_mat[i][j] / num_scored_items
            denominator += d * expected_count / num_scored_items

    return 1.0 - numerator / denominator


def kappa(rater_a, rater_b, min_rating=None, max_rating=None):
    """
    Calculates the kappa
    kappa calculates the kappa
    value, which is a measure of inter-rater agreement between two raters
    that provide discrete numeric ratings.  Potential values range from -1
    (representing complete disagreement) to 1 (representing complete
    agreement).  A kappa value of 0 is expected if all agreement is due to
    chance.
    kappa(rater_a, rater_b), where rater_a and rater_b
    each correspond to a list of integer ratings.  These lists must have the
    same length.
    The ratings should be integers, and it is assumed that they contain
    the complete range of possible ratings.
    kappa(X, min_rating, max_rating), where min_rating
    is the minimum possible rating, and max_rating is the maximum possible
    rating
    """
    assert(len(rater_a) == len(rater_b))
    if min_rating is None:
        min_rating = min(rater_a + rater_b)
    if max_rating is None:
        max_rating = max(rater_a + rater_b)
    conf_mat = confusion_matrix(rater_a, rater_b,
                                min_rating, max_rating)
    num_ratings = len(conf_mat)
    num_scored_items = float(len(rater_a))

    hist_rater_a = histogram(rater_a, min_rating, max_rating)
    hist_rater_b = histogram(rater_b, min_rating, max_rating)

    numerator = 0.0
    denominator = 0.0

    for i in range(num_ratings):
        for j in range(num_ratings):
            expected_count = (hist_rater_a[i] * hist_rater_b[j]
                              / num_scored_items)
            if i == j:
                d = 0.0
            else:
                d = 1.0
            numerator += d * conf_mat[i][j] / num_scored_items
            denominator += d * expected_count / num_scored_items

    return 1.0 - numerator / denominator


def mean_quadratic_weighted_kappa(kappas, weights=None):
    """
    Calculates the mean of the quadratic
    weighted kappas after applying Fisher's r-to-z transform, which is
    approximately a variance-stabilizing transformation.  This
    transformation is undefined if one of the kappas is 1.0, so all kappa
    values are capped in the range (-0.999, 0.999).  The reverse
    transformation is then applied before returning the result.
    mean_quadratic_weighted_kappa(kappas), where kappas is a vector of
    kappa values
    mean_quadratic_weighted_kappa(kappas, weights), where weights is a vector
    of weights that is the same size as kappas.  Weights are applied in the
    z-space
    """
    kappas = np.array(kappas, dtype=float)
    if weights is None:
        weights = np.ones(np.shape(kappas))
    else:
        weights = weights / np.mean(weights)

    # ensure that kappas are in the range [-.999, .999]
    kappas = np.array([min(x, .999) for x in kappas])
    kappas = np.array([max(x, -.999) for x in kappas])

    z = 0.5 * np.log((1 + kappas) / (1 - kappas)) * weights
    z = np.mean(z)
    return (np.exp(2 * z) - 1) / (np.exp(2 * z) + 1)


def weighted_mean_quadratic_weighted_kappa(solution, submission):
    predicted_score = submission[submission.columns[-1]].copy()
    predicted_score.name = "predicted_score"
    if predicted_score.index[0] == 0:
        predicted_score = predicted_score[:len(solution)]
        predicted_score.index = solution.index
    combined = solution.join(predicted_score, how="left")
    groups = combined.groupby(by="essay_set")
    kappas = [quadratic_weighted_kappa(group[1]["essay_score"], group[1]["predicted_score"]) for group in groups]
    weights = [group[1]["essay_weight"].irow(0) for group in groups]
    return mean_quadratic_weighted_kappa(kappas, weights=weights)


# Copyright (c) OpenMMLab. All rights reserved.
import math

import torch
from torch.utils.data import DistributedSampler as _DistributedSampler

from antgo.framework.helper.utils.dist_utils import sync_random_seed
from antgo.framework.helper.utils.util_distribution import get_device

class DistributedSampler(_DistributedSampler):

    def __init__(self,
                 dataset,
                 num_replicas=None,
                 rank=None,
                 shuffle=True,
                 seed=0,strategy=None):
        super().__init__(
            dataset, num_replicas=num_replicas, rank=rank, shuffle=shuffle)

        assert(strategy is None)
        # In distributed sampling, different ranks should sample
        # non-overlapped data in the dataset. Therefore, this function
        # is used to make sure that each rank shuffles the data indices
        # in the same order based on the same seed. Then different ranks
        # could use different indices to select non-overlapped data from the
        # same data list.
        device = get_device()
        self.seed = sync_random_seed(seed, device)

    def __iter__(self):
        # deterministically shuffle based on epoch
        if self.shuffle:
            g = torch.Generator()
            # When :attr:`shuffle=True`, this ensures all replicas
            # use a different random ordering for each epoch.
            # Otherwise, the next iteration of this sampler will
            # yield the same ordering.
            g.manual_seed(self.epoch + self.seed)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = torch.arange(len(self.dataset)).tolist()

        # add extra samples to make it evenly divisible
        # in case that indices is shorter than half of total_size
        indices = (indices *
                   math.ceil(self.total_size / len(indices)))[:self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)
